# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

from typing import Tuple
import torch
from torch import Tensor
import triton
import triton.language as tl
from triton.language.extra import libdevice
from .triton_kernels.utils import IS_HIP, get_num_SMs
from .dtypes import *


# Get dtype min/max range based on compute dtype
def get_dtype_range(compute_dtype: torch.dtype) -> float:
    if compute_dtype.is_floating_point:
        dtype_info = torch.finfo(compute_dtype)
    else:
        dtype_info = torch.iinfo(compute_dtype)
    return dtype_info.min, dtype_info.max


NVFP4_META_SCALE = 0.05  # Temporary NVFP logic
####################################################################################################################
# MXFP4 / NVFP4 weight quantizer
####################################################################################################################

# Cache workspace for multiple gpus (less than a KB per GPU)
fp4_values, fp4_p_vals, fp4_thresholds, thr_pos = [], [], [], []
for g_id in range(torch.cuda.device_count()):
    current_device = "cuda:" + str(g_id)

    fp4_values.append(
        torch.tensor(
            [0, 0.5, 1, 1.5, 2, 3, 4, 6, -0, -0.5, -1, -1.5, -2, -3, -4, -6],
            dtype=torch.float32,
            device=current_device,
        )
    )

    fp4_p_vals.append(
        torch.tensor(
            [0, 0.5, 1, 1.5, 2, 3, 4, 6],
            dtype=torch.float32,
            device=current_device,
        )
    )

    fp4_thresholds.append(
        torch.tensor(
            [0.25, 0.75, 1.25, 1.75, 2.5, 3.5, 5.0],
            dtype=torch.float32,
            device=current_device,
        )
    )  # (fp4_p_vals[:-1] + fp4_p_vals[1:]) / 2

    fp4_pos = torch.tensor(
        [0, 0.5, 1, 1.5, 2, 3, 4, 6],
        dtype=torch.float32,
        device=current_device,
    )

    thr_pos.append(
        # last val is dummy to make len a power of 2
        torch.tensor(
            [0.25, 0.75, 1.25, 1.75, 2.5, 3.5, 5.0, 7.0],
            dtype=torch.float32,
            device=current_device,
        )
    )  # (fp4_p_vals[:-1] + fp4_p_vals[1:]) / 2


class WeightQuantizerMXFP:
    def __init__(self, compute_dtype=torch.bfloat16, device="cuda:0"):
        self.compute_dtype = compute_dtype
        self.device = device

    def round_to_closest_fp4(self, tensor):
        device_index = tensor.device.index
        out = fp4_p_vals[device_index][
            torch.searchsorted(
                fp4_thresholds[device_index].to(tensor.dtype), tensor.abs()
            )
        ].to(tensor.dtype)
        out *= tensor.sign()
        return out

    def to_index(self, W_q):
        assert W_q.is_floating_point(), "Input should be floating point fp4 values."
        device_index = W_q.device.index
        return (
            (W_q.view(-1, 1) == fp4_values[device_index].to(W_q.dtype).view(1, -1))
            .to(torch.uint8)
            .argmax(dim=1)
            .to(torch.uint8)
            .view(W_q.shape)
        )

    @torch.compile(fullgraph=True)
    def quantize_mxfp8(
        self,
        W: torch.Tensor,
        index: bool = False,
        mx_fp8_dtype: torch.dtype = torch.float8_e4m3fn,
    ) -> (torch.Tensor, torch.Tensor):
        group_size: int = 32
        eps_exp: int = -30
        eps: float = 2**eps_exp
        min_val = torch.finfo(mx_fp8_dtype).min
        max_val = torch.finfo(mx_fp8_dtype).max

        W_flat = W.view(-1, group_size).float()
        ideal_scale = W_flat.abs().amax(dim=1, keepdim=True)
        ideal_scale /= max_val

        scales = (2 ** torch.ceil(torch.log2(ideal_scale))).clamp_(min=eps)

        W_q = (W_flat / scales).clamp_(min=min_val, max=max_val)
        scales = scales.to(torch.float8_e8m0fnu)

        if index:
            W_q = W_q.to(mx_fp8_dtype)
        else:
            W_q = W_q.to(mx_fp8_dtype).to(W_flat.dtype)

        return W_q, scales

    @torch.compile(fullgraph=True)
    def quantize_mxfp4(
        self, W: torch.Tensor, window_size: int = 0, index: bool = False
    ) -> (torch.Tensor, torch.Tensor):
        group_size: int = 32
        eps_exp: int = -30
        eps: float = 2**eps_exp
        W_nbits = 4
        max_val = 6

        W_flat = W.view(-1, group_size).float()
        ideal_scale = W_flat.abs().amax(dim=1, keepdim=True)
        ideal_scale /= max_val

        if window_size == 0:
            scales = 2 ** torch.ceil(torch.log2(ideal_scale))
        else:
            initial_log2_scales = torch.ceil(torch.log2(ideal_scale))
            search_offsets = torch.arange(
                -window_size,
                window_size + 1,
                device=W.device,
                dtype=initial_log2_scales.dtype,
            ).view(1, -1)
            candidate_scales = torch.pow(2, initial_log2_scales + search_offsets)
            candidate_scales[candidate_scales < eps] = eps

            W_q_candidates = self.round_to_closest_fp4(
                W_flat.unsqueeze(1) / candidate_scales.unsqueeze(-1)
            )
            W_r_candidates = W_q_candidates * candidate_scales.unsqueeze(-1)
            errors = (W_flat.unsqueeze(1) - W_r_candidates).abs().mean(dim=-1)
            scales = torch.gather(
                candidate_scales, 1, torch.argmin(errors, dim=1, keepdim=True)
            )

        scales = scales.clamp_(eps)
        W_q = self.round_to_closest_fp4(W_flat / scales)
        scales = scales.to(torch.float8_e8m0fnu)

        if index:
            W_q = self.to_index(W_q)
        return W_q, scales

    @torch.compile(fullgraph=True)
    def quantize_nvfp4(
        self, W: torch.Tensor, window_size: int = 0, index: bool = False
    ) -> (torch.Tensor, torch.Tensor):

        group_size: int = 16
        eps: float = 1e-6
        W_nbits = 4
        max_val = 6
        fp8_dtype = torch.float8_e4m3fn  # This is for Nvidia only.
        max_fp8 = torch.finfo(fp8_dtype).max  # 448

        W_flat = W.view(-1, group_size).float()
        ideal_scale = W_flat.abs().amax(dim=1, keepdim=True)
        ideal_scale /= max_val
        meta_scales = (
            NVFP4_META_SCALE  # ideal_scale.max().clamp_(min=eps) - TODO: use max()
        )
        ideal_scale /= meta_scales
        ideal_scale = ideal_scale.clamp_(max=max_fp8).to(fp8_dtype)

        if window_size == 0:
            scales = ideal_scale
        else:
            search_offsets = torch.arange(
                -window_size, window_size + 1, device=W.device, dtype=torch.int
            ).view(1, -1)

            candidate_scales = (
                (ideal_scale.view(torch.int8) + search_offsets)
                .clamp_(-128, 127)
                .to(torch.int8)
            )

            # Avoid nan in int8 range (-1, 127 as int8 as e4m3 nans)
            candidate_scales[candidate_scales == -1] = 1
            candidate_scales[candidate_scales == 127] = 1
            candidate_scales = candidate_scales.view(fp8_dtype).float()
            candidate_scales[candidate_scales < eps] = eps

            W_q_candidates = self.round_to_closest_fp4(
                W_flat.unsqueeze(1) / (candidate_scales * meta_scales).unsqueeze(-1)
            )
            W_r_candidates = W_q_candidates * candidate_scales.unsqueeze(-1)
            errors = (W_flat.unsqueeze(1) - W_r_candidates).abs().mean(dim=-1)
            scales = torch.gather(
                candidate_scales, 1, torch.argmin(errors, dim=1, keepdim=True)
            ).to(fp8_dtype)

        scales_full = (scales.to(W_flat.dtype) * meta_scales).clamp_(min=eps)
        W_q = self.round_to_closest_fp4(W_flat / scales_full)

        if index:
            W_q = self.to_index(W_q)

        return W_q, scales

    def dequantize(self, W_q, scales, shape=None, dtype=None):
        if W_q.dtype == torch.uint8:  # from indices
            device_index = W_q.device.index
            W_q = fp4_values[device_index][W_q.int()]

        group_size = W_q.numel() // scales.numel()
        out = W_q.view([-1, group_size]).float() * scales.float()
        if shape is not None:
            out = out.view(shape)
        return out.to(self.compute_dtype if dtype is None else dtype)


####################################################################################################################
# INT8 / FP8 activations
####################################################################################################################
# Main activation scaling functions
@torch.compile(fullgraph=True)
def scale_activations_per_token_torch(
    tensor: Tensor, w_dtype: torch.dtype, fp32_scale: bool = True
) -> Tuple[Tensor, Tensor]:

    min_val, max_val = get_dtype_range(w_dtype)
    if fp32_scale:
        tensor = tensor.to(torch.float32, copy=False)
    out_shape = tensor.shape
    out = tensor.view(-1, tensor.shape[-1])
    scales = torch.abs(out).amax(axis=1, keepdim=True)
    # if(fp32_scale):
    #     scales = scales.to(torch.float32)
    scales.div_(max_val)
    scales.clamp_(min=1e-6)
    out = tensor / scales
    out.clamp_(min_val, max_val)

    if not w_dtype.is_floating_point:
        out.round_()

    out = out.to(dtype=w_dtype)
    return out.view(out_shape), scales


@triton.jit
def round_triton_nvidia(tensor):
    return libdevice.round(tensor)


@triton.jit
def round_triton_amd(tensor):
    return libdevice.floor(tensor + 0.50)


if IS_HIP:
    round_triton = round_triton_amd
else:
    round_triton = round_triton_nvidia


@triton.jit
def scale_activations_per_token_kernel(
    tensor_ptr,
    scale_ptr,
    y_ptr,
    M,
    K,
    stride_m,
    stride_k,
    stride_sm,
    ROUND: tl.constexpr,
    UNROLL: tl.constexpr,
    min_val: tl.constexpr,
    max_val: tl.constexpr,
    fp32_scale: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_K: tl.constexpr,
):
    pid_m = tl.program_id(0) * UNROLL
    pid_k = tl.program_id(1)

    offs_k = pid_k * BLOCK_K + tl.arange(0, BLOCK_K)
    offs_m = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)

    for m in range(UNROLL):
        mask = ((offs_m < M)[:, None] & (offs_k < K)[None, :]).to(tl.int1)
        in_ptrs = offs_m[:, None] * stride_m + offs_k[None, :] * stride_k
        tensor = tl.load(tensor_ptr + in_ptrs, mask=mask, other=0.0)
        if fp32_scale:
            tensor = tensor.to(tl.float32)

        scales_x = tl.max(tl.abs(tensor), axis=1, keep_dims=True)
        scales_x /= max_val
        scales_x = tl.maximum(scales_x, 1e-6)
        tensor /= scales_x
        tensor = tl.minimum(tl.maximum(tensor, min_val), max_val)

        if ROUND:
            tensor = round_triton(tensor)

        tl.store(scale_ptr + offs_m[:, None] * stride_sm, scales_x)
        tl.store(y_ptr + in_ptrs, tensor, mask=mask)
        offs_m += BLOCK_M


def scale_activations_per_token_triton(
    tensor: Tensor, w_dtype: torch.dtype, fp32_scale: bool = True
) -> Tuple[Tensor, Tensor]:
    min_val, max_val = get_dtype_range(w_dtype)
    x_shape = tensor.shape
    tensor = tensor.view(-1, tensor.shape[-1])
    M, K = tensor.shape
    scales = torch.empty(
        (M, 1),
        dtype=torch.float32 if fp32_scale else tensor.dtype,
        device=tensor.device,
    )
    y = torch.empty((M, K), dtype=w_dtype, device=tensor.device)

    UNROLL = 1  # max(1, M // 128)
    BLOCK_M = 1
    BLOCK_K = triton.next_power_of_2(K)
    grid = (triton.cdiv(M, BLOCK_M * UNROLL), triton.cdiv(K, BLOCK_K))

    ROUND = not w_dtype.is_floating_point

    scale_activations_per_token_kernel[grid](
        tensor,
        scales,
        y,
        M,
        K,
        tensor.stride(0),
        tensor.stride(1),
        scales.stride(0),
        min_val=min_val,
        max_val=max_val,
        fp32_scale=fp32_scale,
        ROUND=ROUND,
        UNROLL=UNROLL,
        BLOCK_M=BLOCK_M,
        BLOCK_K=BLOCK_K,
        num_stages=1,
        num_warps=4,
    )

    return y.view(x_shape), scales


####################################################################################################################
# MXFP8
####################################################################################################################
@triton.jit
def next_power_of_2_log_triton(val, eps: tl.constexpr):
    exp = tl.ceil(tl.log2(val)).to(tl.int32)
    exp = tl.maximum(tl.minimum(exp, 254), 127 + eps_exp)
    scales = tl.where(exp >= 0, 1 << scales_log2, 1.0 / (1 << (-exp)))
    return scales, exp


@triton.jit
def next_power_of_2_logapprox_triton(val, eps_exp: tl.constexpr):
    exp = tl.inline_asm_elementwise(
        """
        {
        lg2.approx.f32 $1, $1;
        cvt.rpi.f32.f32 $1, $1;
        cvt.rzi.s32.f32 $0, $1;
        }
        """,
        "=r,r",
        [val],
        dtype=tl.int32,
        is_pure=True,
        pack=1,
    )

    exp = tl.maximum(tl.minimum(exp, 254), 127 + eps_exp)
    scales = tl.where(exp >= 0, 1 << exp, 1.0 / (1 << (-exp)))
    return scales, exp


@triton.jit
def next_power_of_2_bitwise_triton(val, eps_exp: tl.constexpr):
    xi = tl.cast(val, tl.uint32, bitcast=True)
    exp = (xi >> 23) & 0xFF
    mant = xi & 0x7FFFFF
    exp += tl.where(mant != 0, 1, 0)
    exp = tl.maximum(tl.minimum(exp, 254), 127 + eps_exp)
    yi = exp << 23
    scales = tl.cast(yi, tl.float32, bitcast=True)
    return scales, exp


next_power_of_2_triton = next_power_of_2_bitwise_triton


@torch.compile(fullgraph=True)
def scale_activations_mxfp8_torch(
    tensor: Tensor, w_dtype: torch.dtype = torch.float8_e4m3fn
) -> Tuple[Tensor, Tensor]:

    group_size: int = 32
    eps_exp: int = -30
    eps: float = 2**eps_exp
    min_val, max_val = get_dtype_range(w_dtype)

    orig_shape = tensor.shape
    tensor = tensor.view(-1, tensor.shape[-1])
    inter_shape = tensor.shape

    pad_rows = (group_size - inter_shape[0] % group_size) % group_size
    if pad_rows > 0:
        tensor = torch.nn.functional.pad(tensor, (0, 0, 0, pad_rows))
    post_pad_shape = tensor.shape

    W_flat = tensor.view(-1, group_size).float()
    scales = W_flat.abs().amax(dim=1, keepdim=True)
    scales /= max_val
    scales = (2 ** torch.ceil(torch.log2(scales))).clamp_(eps)

    W_q = (W_flat / scales).clamp_(min_val, max_val).to(w_dtype)
    if pad_rows > 0:
        W_q = W_q.view(post_pad_shape)[: inter_shape[0], :]

    W_q = W_q.view(orig_shape)
    scales = (
        scales.to(torch.float8_e8m0fnu)
        .view(torch.uint8)
        .view(post_pad_shape[0], post_pad_shape[1] // group_size)
    )

    return W_q, scales


@triton.jit
def scale_activations_mxfp8_triton_v1_kernel(
    tensor_ptr,
    out_ptr,
    scales_ptr,
    E,
    min_val: tl.constexpr,
    max_val: tl.constexpr,
    eps_exp: tl.constexpr,
    UNROLL: tl.constexpr,
    GROUP_SIZE: tl.constexpr,
):
    pid = tl.program_id(axis=0) * UNROLL

    for m in range(UNROLL):
        offs = pid * GROUP_SIZE + tl.arange(0, GROUP_SIZE)
        mask = (offs < E).to(tl.int1)
        tensor = tl.load(tensor_ptr + offs, mask=mask, other=0.0).to(tl.float32)

        scales, scales_log2 = next_power_of_2_triton(
            tl.max(tl.abs(tensor)) / max_val, eps_exp
        )

        out = tensor / scales
        out = tl.clamp(out, min=min_val, max=max_val)
        out = out.to(out_ptr.dtype.element_ty)
        tl.store(out_ptr + offs, out)
        tl.store(scales_ptr + pid, scales_log2)

        pid += 1


def scale_activations_mxfp8_triton_v1(
    tensor: torch.Tensor, w_dtype: torch.dtype = torch.float8_e4m3fn
) -> Tuple[torch.Tensor, torch.Tensor]:
    group_size: int = 32
    eps_exp: int = -30
    eps: float = 2**eps_exp
    min_val, max_val = get_dtype_range(w_dtype)
    tensor = tensor.contiguous()

    orig_shape = tensor.shape
    tensor = tensor.view(-1, tensor.shape[-1])
    inter_shape = tensor.shape
    pad_rows = (group_size - inter_shape[0] % group_size) % group_size
    post_pad_shape = (inter_shape[0] + pad_rows, inter_shape[1])
    E = tensor.numel()

    UNROLL = min(triton.cdiv(triton.cdiv(E, group_size), get_num_SMs(tensor.device)), 1)

    out = torch.empty(inter_shape, device=tensor.device, dtype=w_dtype)

    scales = torch.empty(
        (post_pad_shape[0], post_pad_shape[1] // group_size),
        device=tensor.device,
        dtype=torch.uint8,
    )

    grid = lambda meta: (triton.cdiv(E // UNROLL, group_size),)
    scale_activations_mxfp8_triton_v1_kernel[grid](
        tensor,
        out,
        scales,
        E=E,
        min_val=min_val,
        max_val=max_val,
        eps_exp=eps_exp,
        UNROLL=UNROLL,
        GROUP_SIZE=group_size,
        num_stages=1,
        num_warps=4,
    )

    return out.view(orig_shape), scales


@triton.jit
def scale_activations_mxfp8_triton_kernel_v2(
    tensor_ptr,
    out_ptr,
    scales_ptr,
    M,
    K,
    stride_m_t,
    stride_k_t,
    stride_m_s,
    stride_k_s,
    stride_m_o,
    stride_k_o,
    #########################
    min_val: tl.constexpr,
    max_val: tl.constexpr,
    eps_exp: tl.constexpr,
    GROUP_SIZE: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr,
):
    pid_m = tl.program_id(axis=0)
    pid_k = tl.program_id(axis=1)
    out_dtype: tl.constexpr = out_ptr.dtype.element_ty

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_k = pid_k * GROUP_SIZE + tl.arange(0, GROUP_SIZE)

    # Load
    mask = ((offs_m[:, None] < M) & (offs_k[None, :] < K)).to(tl.int1)
    tensor_ptrs = tensor_ptr + (
        offs_m[:, None] * stride_m_t + offs_k[None, :] * stride_k_t
    )
    tensor = tl.load(tensor_ptrs, mask=mask, other=0.0).to(tl.float32)

    # next power of 2 via log
    scales, scales_log2 = next_power_of_2_triton(
        tl.max(tl.abs(tensor), axis=1, keep_dims=True) / max_val, eps_exp
    )

    # Map to index
    out = tensor / scales
    out = tl.clamp(out, min=min_val, max=max_val)
    out = out.to(out_dtype)

    # Store
    out_mask = ((offs_m[:, None] < M) & (offs_k[None, :] < K)).to(tl.int1)
    tl.store(
        out_ptr + (offs_m[:, None] * stride_m_o + offs_k[None, :] * stride_k_o),
        out,
        mask=out_mask,
    )

    offs_k = pid_k * 1 + tl.arange(0, 1)
    tl.store(
        scales_ptr + (offs_m[:, None] * stride_m_s + offs_k[None, :] * stride_k_s),
        scales_log2,
    )


def scale_activations_mxfp8_triton_v2(
    tensor: torch.Tensor, w_dtype: torch.dtype = torch.float8_e4m3fn
) -> Tuple[torch.Tensor, torch.Tensor]:
    group_size: int = 32
    eps_exp: int = -30
    eps: float = 2**-30
    min_val, max_val = get_dtype_range(w_dtype)

    tensor = tensor.contiguous()
    tensor = tensor.view(-1, tensor.shape[-1])
    M, K = tensor.shape

    pad_m = (group_size - M % group_size) % group_size
    M_padded = M + pad_m

    out = torch.empty((M, K), device=tensor.device, dtype=w_dtype)
    scales = torch.empty(
        (M_padded, K // group_size), device=tensor.device, dtype=torch.uint8
    )

    # BLOCK_SIZE_M = min(max(next_power_of_2(M), group_size), 128)
    BLOCK_SIZE_M = group_size
    grid = (triton.cdiv(M, BLOCK_SIZE_M), triton.cdiv(K, group_size))
    device_index = tensor.device.index

    scale_activations_mxfp8_triton_kernel_v2[grid](
        tensor,
        out,
        scales,
        M,
        K,
        tensor.stride(0),
        tensor.stride(1),
        scales.stride(0),
        scales.stride(1),
        out.stride(0),
        out.stride(1),
        #########################
        min_val=min_val,
        max_val=max_val,
        eps_exp=eps_exp,
        GROUP_SIZE=group_size,
        BLOCK_SIZE_M=BLOCK_SIZE_M,
        num_stages=2,
        num_warps=4,
    )

    return out, scales


####################################################################################################################
# MXPF4 / NVFP4
####################################################################################################################
@torch.compile(fullgraph=True)
def scale_activations_mxfp4_torch(tensor: Tensor) -> Tuple[Tensor, Tensor]:
    group_size: int = 32
    eps_exp: int = -30
    eps: float = 2**eps_exp
    max_val: float = 6

    orig_shape = tensor.shape
    tensor = tensor.view(-1, tensor.shape[-1])
    inter_shape = tensor.shape

    pad_rows = (group_size - inter_shape[0] % group_size) % group_size
    if pad_rows > 0:
        tensor = torch.nn.functional.pad(tensor, (0, 0, 0, pad_rows))
    post_pad_shape = tensor.shape

    W_flat = tensor.view(-1, group_size).float()
    scales = W_flat.abs().amax(dim=1, keepdim=True)
    scales /= max_val
    scales = (2 ** torch.ceil(torch.log2(scales))).clamp_(eps)

    W_q = W_flat / scales
    if pad_rows > 0:
        W_q = W_q.view(post_pad_shape)[: inter_shape[0], :]

    # 1) Map to closest index
    device_index = W_q.device.index

    W_q = (
        (W_q.view(-1, 1) - fp4_values[device_index].to(W_q.dtype).view(1, -1))
        .abs()
        .argmin(dim=1)
        .to(torch.uint8)
        .view(inter_shape)
    )
    # 2) Pack
    W_q = (W_q[:, ::2] | W_q[:, 1::2] << 4).to(torch.uint8)

    # Reshape scales
    scales = (
        scales.to(torch.float8_e8m0fnu)
        .view(torch.uint8)
        .view(post_pad_shape[0], post_pad_shape[1] // group_size)
    )
    return W_q, scales


@torch.compile(fullgraph=True)
def scale_activations_nvfp4_torch(tensor: Tensor) -> Tuple[Tensor, Tensor]:
    group_size: int = 16
    eps: float = 1e-6
    max_val: float = 6
    fp8_dtype = torch.float8_e4m3fn  # Support Nvidia only
    max_fp8 = torch.finfo(fp8_dtype).max  # 448

    orig_shape = tensor.shape
    tensor = tensor.view(-1, tensor.shape[-1])
    inter_shape = tensor.shape

    pad_rows = (group_size - inter_shape[0] % group_size) % group_size
    if pad_rows > 0:
        tensor = torch.nn.functional.pad(tensor, (0, 0, 0, pad_rows))
    post_pad_shape = tensor.shape

    W_flat = tensor.view(-1, group_size).float()
    scales = W_flat.abs().amax(dim=1, keepdim=True)
    scales /= max_val
    meta_scales = NVFP4_META_SCALE  # scales.max().clamp_(min=eps) - TODO: use max()
    scales /= meta_scales
    scales = scales.clamp(max=max_fp8).to(fp8_dtype).to(W_flat.dtype)

    W_q = W_flat / (scales * meta_scales)
    if pad_rows > 0:
        W_q = W_q.view(post_pad_shape)[: inter_shape[0], :]

    # 1) Map to closest index
    device_index = W_q.device.index

    W_q = (
        (W_q.view(-1, 1) - fp4_values[device_index].to(W_q.dtype).view(1, -1))
        .abs()
        .argmin(dim=1)
        .to(torch.uint8)
        .view(inter_shape)
    )
    # 2) Pack
    W_q = (W_q[:, ::2] | W_q[:, 1::2] << 4).to(torch.uint8)

    # Reshape scales
    scales = scales.to(fp8_dtype).view(
        post_pad_shape[0], post_pad_shape[1] // group_size
    )
    return W_q, scales


@triton.jit
def scale_activations_mxfp4_triton_kernel_v1(
    tensor_ptr,
    out_ptr,
    scales_ptr,
    thr_pos_ptr,
    E,
    eps_exp: tl.constexpr,
    UNROLL: tl.constexpr,
    GROUP_SIZE: tl.constexpr,
):
    pid = tl.program_id(axis=0) * UNROLL

    HALF_GROUP_SIZE: tl.constexpr = GROUP_SIZE // 2
    out_dtype: tl.constexpr = out_ptr.dtype.element_ty
    thr_pos = tl.load(thr_pos_ptr + tl.arange(0, 8), eviction_policy="evict_last")[
        None, :
    ]

    for m in range(UNROLL):
        # Load
        offs = pid * GROUP_SIZE + tl.arange(0, GROUP_SIZE)
        mask = (offs < E).to(tl.int1)
        tensor = tl.load(tensor_ptr + offs, mask=mask, other=0.0).to(tl.float32)

        scales, scales_log2 = next_power_of_2_triton(
            tl.max(tl.abs(tensor)) / 6.0, eps_exp
        )

        # Map to index
        wq = tensor / scales
        idx_abs = tl.sum(tl.abs(wq[:, None]) > thr_pos, axis=1)
        out = tl.where(wq >= 0, idx_abs, idx_abs + 8).to(out_dtype)

        # Pack
        lo, hi = tl.split(out.reshape((HALF_GROUP_SIZE, 2), can_reorder=False))
        out = lo | (hi << 4)

        # Store
        offs_out = pid * HALF_GROUP_SIZE + tl.arange(0, HALF_GROUP_SIZE)
        tl.store(out_ptr + offs_out, out)
        tl.store(scales_ptr + pid, scales_log2)

        pid += 1


def scale_activations_mxfp4_triton_v1(tensor: Tensor) -> Tuple[Tensor, Tensor]:
    group_size: int = 32
    eps_exp: int = -30
    eps = 2**eps_exp
    tensor = tensor.contiguous()

    orig_shape = tensor.shape
    tensor = tensor.view(-1, tensor.shape[-1])
    inter_shape = (tensor.shape[0], tensor.shape[1] // 2)
    pad_rows = (group_size - inter_shape[0] % group_size) % group_size
    post_pad_shape = (inter_shape[0] + pad_rows, inter_shape[1])
    E = tensor.numel()

    UNROLL = min(triton.cdiv(triton.cdiv(E, group_size), get_num_SMs(tensor.device)), 1)

    out = torch.empty(inter_shape, device=tensor.device, dtype=torch.uint8)
    scales = torch.empty(
        (post_pad_shape[0], post_pad_shape[1] * 2 // group_size),
        device=tensor.device,
        dtype=torch.uint8,
    )
    device_index = tensor.device.index

    grid = lambda meta: (triton.cdiv(E // UNROLL, group_size),)
    scale_activations_mxfp4_triton_kernel_v1[grid](
        tensor,
        out,
        scales,
        thr_pos[device_index],
        E,
        eps_exp=eps_exp,
        UNROLL=UNROLL,
        GROUP_SIZE=group_size,
        num_stages=1,
        num_warps=4,
    )

    return out, scales


@triton.jit
def scale_activations_mxfp4_triton_kernel_v2(
    tensor_ptr,
    out_ptr,
    scales_ptr,
    thr_pos_ptr,
    M,
    K,
    stride_m_t,
    stride_k_t,
    stride_m_s,
    stride_k_s,
    stride_m_o,
    stride_k_o,
    #########################
    eps_exp: tl.constexpr,
    GROUP_SIZE: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr,
):
    pid_m = tl.program_id(axis=0)
    pid_k = tl.program_id(axis=1)

    HALF_GROUP_SIZE: tl.constexpr = GROUP_SIZE // 2
    out_dtype: tl.constexpr = out_ptr.dtype.element_ty
    thr_pos = tl.load(thr_pos_ptr + tl.arange(0, 8), eviction_policy="evict_last")[
        None, :
    ]

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_k = pid_k * GROUP_SIZE + tl.arange(0, GROUP_SIZE)

    # Load
    mask = ((offs_m[:, None] < M) & (offs_k[None, :] < K)).to(tl.int1)
    tensor_ptrs = tensor_ptr + (
        offs_m[:, None] * stride_m_t + offs_k[None, :] * stride_k_t
    )
    tensor = tl.load(tensor_ptrs, mask=mask, other=0.0).to(tl.float32)

    # next power of 2 via log
    scales, scales_log2 = next_power_of_2_triton(
        tl.max(tl.abs(tensor), axis=1, keep_dims=True) / 6.0, eps_exp
    )

    # Map to index
    wq = tensor / scales
    idx_abs = tl.sum(tl.abs(wq[:, :, None]) > thr_pos[None, :, :], axis=2)
    out = tl.where(wq >= 0, idx_abs, idx_abs + 8).to(out_dtype)

    # Pack
    lo, hi = tl.split(
        out.reshape((BLOCK_SIZE_M, HALF_GROUP_SIZE, 2), can_reorder=False)
    )
    out = lo | (hi << 4)

    # Store
    offs_k = pid_k * HALF_GROUP_SIZE + tl.arange(0, HALF_GROUP_SIZE)
    out_mask = ((offs_m[:, None] < M) & (offs_k[None, :] < (K // 2))).to(tl.int1)
    tl.store(
        out_ptr + (offs_m[:, None] * stride_m_o + offs_k[None, :] * stride_k_o),
        out,
        mask=out_mask,
    )

    offs_k = pid_k * 1 + tl.arange(0, 1)
    tl.store(
        scales_ptr + (offs_m[:, None] * stride_m_s + offs_k[None, :] * stride_k_s),
        scales_log2,
    )


def scale_activations_mxfp4_triton_v2(tensor: Tensor) -> Tuple[Tensor, Tensor]:
    group_size: int = 32
    eps_exp: int = -30
    eps: float = 2**eps_exp

    tensor = tensor.contiguous()
    tensor = tensor.view(-1, tensor.shape[-1])
    M, K = tensor.shape

    pad_m = (group_size - M % group_size) % group_size
    M_padded = M + pad_m

    out = torch.empty((M, K // 2), device=tensor.device, dtype=torch.uint8)
    scales = torch.empty(
        (M_padded, K // group_size), device=tensor.device, dtype=torch.uint8
    )

    # BLOCK_SIZE_M = min(max(next_power_of_2(M), group_size), 128)
    BLOCK_SIZE_M = group_size

    grid = (triton.cdiv(M, BLOCK_SIZE_M), triton.cdiv(K, group_size))
    device_index = tensor.device.index

    scale_activations_mxfp4_triton_kernel_v2[grid](
        tensor,
        out,
        scales,
        thr_pos[device_index],
        M,
        K,
        tensor.stride(0),
        tensor.stride(1),
        scales.stride(0),
        scales.stride(1),
        out.stride(0),
        out.stride(1),
        #########################
        eps_exp=eps_exp,
        GROUP_SIZE=group_size,
        BLOCK_SIZE_M=BLOCK_SIZE_M,
        num_stages=2,
        num_warps=4,
    )

    return out, scales


@triton.jit
def scale_activations_nvfp4_triton_kernel_v2(
    tensor_ptr,
    out_ptr,
    scales_ptr,
    thr_pos_ptr,
    M,
    K,
    stride_m_t,
    stride_k_t,
    stride_m_s,
    stride_k_s,
    stride_m_o,
    stride_k_o,
    #########################
    eps: tl.constexpr,
    GROUP_SIZE: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr,
    meta_scales: tl.constexpr = NVFP4_META_SCALE,
):

    pid_m = tl.program_id(axis=0)
    pid_k = tl.program_id(axis=1)

    fp8_dtype: tl.constexpr = tl.float8e4nv
    max_fp8: tl.constexpr = 448.0
    HALF_GROUP_SIZE: tl.constexpr = GROUP_SIZE // 2
    out_dtype: tl.constexpr = out_ptr.dtype.element_ty
    thr_pos = tl.load(thr_pos_ptr + tl.arange(0, 8), eviction_policy="evict_last")[
        None, :
    ]
    # thr_pos += 1e-6

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_k = pid_k * GROUP_SIZE + tl.arange(0, GROUP_SIZE)

    # Load
    mask = ((offs_m[:, None] < M) & (offs_k[None, :] < K)).to(tl.int1)
    tensor_ptrs = tensor_ptr + (
        offs_m[:, None] * stride_m_t + offs_k[None, :] * stride_k_t
    )
    tensor = tl.load(tensor_ptrs, mask=mask, other=0.0).to(tl.float32)

    # FP8 scales
    scales = tl.max(tl.abs(tensor), axis=1, keep_dims=True) / (6.0 * meta_scales)
    scales = tl.minimum(scales, max_fp8).to(fp8_dtype)

    # Map to index
    scales_full = tl.maximum(scales.to(tl.float32) * meta_scales, eps)
    wq = tensor / scales_full
    idx_abs = tl.sum(tl.abs(wq[:, :, None]) > thr_pos[None, :, :], axis=2)
    out = tl.where(wq >= 0, idx_abs, idx_abs + 8).to(out_dtype)

    # Pack
    lo, hi = tl.split(
        out.reshape((BLOCK_SIZE_M, HALF_GROUP_SIZE, 2), can_reorder=False)
    )
    out = lo | (hi << 4)

    # Store
    offs_k = pid_k * HALF_GROUP_SIZE + tl.arange(0, HALF_GROUP_SIZE)
    out_mask = ((offs_m[:, None] < M) & (offs_k[None, :] < (K // 2))).to(tl.int1)
    tl.store(
        out_ptr + (offs_m[:, None] * stride_m_o + offs_k[None, :] * stride_k_o),
        out,
        mask=out_mask,
    )

    offs_k = pid_k + tl.arange(0, 1)
    tl.store(
        scales_ptr + (offs_m[:, None] * stride_m_s + offs_k[None, :] * stride_k_s),
        scales,
    )


def scale_activations_nvfp4_triton_v2(
    tensor: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    group_size: int = 16
    eps: float = 1e-6
    fp8_dtype = torch.float8_e4m3fn  # Nvidia only

    tensor = tensor.contiguous()
    tensor = tensor.view(-1, tensor.shape[-1])
    M, K = tensor.shape

    pad_m = (group_size - M % group_size) % group_size
    M_padded = M + pad_m

    out = torch.empty((M, K // 2), device=tensor.device, dtype=torch.uint8)
    scales = torch.empty(
        (M_padded, K // group_size), device=tensor.device, dtype=fp8_dtype
    )

    # BLOCK_SIZE_M = min(max(next_power_of_2(M), group_size), 128)
    BLOCK_SIZE_M = group_size
    grid = (triton.cdiv(M, BLOCK_SIZE_M), triton.cdiv(K, group_size))
    device_index = tensor.device.index

    scale_activations_nvfp4_triton_kernel_v2[grid](
        tensor,
        out,
        scales,
        thr_pos[device_index],
        M,
        K,
        tensor.stride(0),
        tensor.stride(1),
        scales.stride(0),
        scales.stride(1),
        out.stride(0),
        out.stride(1),
        #########################
        eps=eps,
        GROUP_SIZE=group_size,
        BLOCK_SIZE_M=BLOCK_SIZE_M,
        num_stages=2,
        num_warps=4,
    )

    return out, scales


####################################################################################################################
scale_activations_per_token = scale_activations_per_token_triton
scale_activations_mxfp8 = scale_activations_mxfp8_triton_v2
scale_activations_mxfp4 = scale_activations_mxfp4_triton_v2
scale_activations_nvfp4 = scale_activations_nvfp4_triton_v2



# Modified from transformers.models.xlm_roberta.modeling_xlm_roberta
# Copyright 2024-2025 The Alibaba Wan Team Authors. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["XLMRoberta", "xlm_roberta_large"]


class SelfAttention(nn.Module):
    def __init__(self, dim, num_heads, dropout=0.1, eps=1e-5):
        assert dim % num_heads == 0
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.eps = eps

        # layers
        self.q = nn.Linear(dim, dim)
        self.k = nn.Linear(dim, dim)
        self.v = nn.Linear(dim, dim)
        self.o = nn.Linear(dim, dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, mask):
        """
        x:   [B, L, C].
        """
        b, s, c, n, d = *x.size(), self.num_heads, self.head_dim

        # compute query, key, value
        q = self.q(x).reshape(b, s, n, d).permute(0, 2, 1, 3)
        k = self.k(x).reshape(b, s, n, d).permute(0, 2, 1, 3)
        v = self.v(x).reshape(b, s, n, d).permute(0, 2, 1, 3)

        # compute attention
        p = self.dropout.p if self.training else 0.0
        x = F.scaled_dot_product_attention(q, k, v, mask, p)
        x = x.permute(0, 2, 1, 3).reshape(b, s, c)

        # output
        x = self.o(x)
        x = self.dropout(x)
        return x


class AttentionBlock(nn.Module):
    def __init__(self, dim, num_heads, post_norm, dropout=0.1, eps=1e-5):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.post_norm = post_norm
        self.eps = eps

        # layers
        self.attn = SelfAttention(dim, num_heads, dropout, eps)
        self.norm1 = nn.LayerNorm(dim, eps=eps)
        self.ffn = nn.Sequential(
            nn.Linear(dim, dim * 4),
            nn.GELU(),
            nn.Linear(dim * 4, dim),
            nn.Dropout(dropout),
        )
        self.norm2 = nn.LayerNorm(dim, eps=eps)

    def forward(self, x, mask):
        if self.post_norm:
            x = self.norm1(x + self.attn(x, mask))
            x = self.norm2(x + self.ffn(x))
        else:
            x = x + self.attn(self.norm1(x), mask)
            x = x + self.ffn(self.norm2(x))
        return x


class XLMRoberta(nn.Module):
    """
    XLMRobertaModel with no pooler and no LM head.
    """

    def __init__(
        self,
        vocab_size=250002,
        max_seq_len=514,
        type_size=1,
        pad_id=1,
        dim=1024,
        num_heads=16,
        num_layers=24,
        post_norm=True,
        dropout=0.1,
        eps=1e-5,
    ):
        super().__init__()
        self.vocab_size = vocab_size
        self.max_seq_len = max_seq_len
        self.type_size = type_size
        self.pad_id = pad_id
        self.dim = dim
        self.num_heads = num_heads
        self.num_layers = num_layers
        self.post_norm = post_norm
        self.eps = eps

        # embeddings
        self.token_embedding = nn.Embedding(vocab_size, dim, padding_idx=pad_id)
        self.type_embedding = nn.Embedding(type_size, dim)
        self.pos_embedding = nn.Embedding(max_seq_len, dim, padding_idx=pad_id)
        self.dropout = nn.Dropout(dropout)

        # blocks
        self.blocks = nn.ModuleList(
            [
                AttentionBlock(dim, num_heads, post_norm, dropout, eps)
                for _ in range(num_layers)
            ]
        )

        # norm layer
        self.norm = nn.LayerNorm(dim, eps=eps)

    def forward(self, ids):
        """
        ids: [B, L] of torch.LongTensor.
        """
        b, s = ids.shape
        mask = ids.ne(self.pad_id).long()

        # embeddings
        x = (
            self.token_embedding(ids)
            + self.type_embedding(torch.zeros_like(ids))
            + self.pos_embedding(self.pad_id + torch.cumsum(mask, dim=1) * mask)
        )
        if self.post_norm:
            x = self.norm(x)
        x = self.dropout(x)

        # blocks
        mask = torch.where(mask.view(b, 1, 1, s).gt(0), 0.0, torch.finfo(x.dtype).min)
        for block in self.blocks:
            x = block(x, mask)

        # output
        if not self.post_norm:
            x = self.norm(x)
        return x


def xlm_roberta_large(pretrained=False, return_tokenizer=False, device="cpu", **kwargs):
    """
    XLMRobertaLarge adapted from Huggingface.
    """
    # params
    cfg = dict(
        vocab_size=250002,
        max_seq_len=514,
        type_size=1,
        pad_id=1,
        dim=1024,
        num_heads=16,
        num_layers=24,
        post_norm=True,
        dropout=0.1,
        eps=1e-5,
    )
    cfg.update(**kwargs)

    # init a model on device
    with torch.device(device):
        model = XLMRoberta(**cfg)
    return model



from typing import Optional, Union, Tuple
from functools import partial

import torch
from torch import Tensor
import triton
import triton.language as tl
from einops import rearrange


def sinusoidal_embedding_1d(dim, position):
    # preprocess
    assert dim % 2 == 0
    half = dim // 2
    position = position.type(torch.float64)

    # calculation
    sinusoid = torch.outer(
        position, torch.pow(10000, -torch.arange(half).to(position).div(half))
    )
    x = torch.cat([torch.cos(sinusoid), torch.sin(sinusoid)], dim=1)
    return x


def fast_rope_apply(x: torch.Tensor, freqs: torch.Tensor) -> torch.Tensor:
    """
    Optimized version of rope_apply using flash_attention's rotary embedding implementation.
    This version processes the entire batch at once for efficiency.

    Args:
        x (Tensor): Input tensor with shape [batch_size, seq_len, n_heads, head_dim]
        freqs (Tensor): Complex frequencies with shape [max_seq_len, head_dim // 2]

    Returns:
        Tensor: Rotary-embedded tensor with same shape as input
    """
    batch_size, seq_len, n_heads, head_dim = x.shape

    # freqs is already sharded to local seq_len under flattened CP
    freqs = freqs.view(seq_len, head_dim // 2)
    cos = torch.cos(freqs).to(torch.float32)
    sin = torch.sin(freqs).to(torch.float32)

    # Apply the rotation
    rotated = apply_rotary_emb(
        x.to(torch.float32), cos, sin, interleaved=True, inplace=False
    )

    return rotated.to(x.dtype)


def rotate_half(x, interleaved=False):
    if not interleaved:
        x1, x2 = x.chunk(2, dim=-1)
        return torch.cat((-x2, x1), dim=-1)
    else:
        x1, x2 = x[..., ::2], x[..., 1::2]
        return rearrange(
            torch.stack((-x2, x1), dim=-1), "... d two -> ... (d two)", two=2
        )


class ApplyRotaryEmb(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        x,
        cos,
        sin,
        interleaved=False,
        inplace=False,
        seqlen_offsets: Union[int, Tensor] = 0,
        cu_seqlens: Optional[Tensor] = None,
        max_seqlen: Optional[int] = None,
    ):
        out = apply_rotary(
            x,
            cos,
            sin,
            seqlen_offsets=seqlen_offsets,
            cu_seqlens=cu_seqlens,
            max_seqlen=max_seqlen,
            interleaved=interleaved,
            inplace=inplace,
        )
        if isinstance(seqlen_offsets, int):
            ctx.save_for_backward(
                cos, sin, cu_seqlens
            )  # Can't save int with save_for_backward
            ctx.seqlen_offsets = seqlen_offsets
        else:
            ctx.save_for_backward(cos, sin, cu_seqlens, seqlen_offsets)
            ctx.seqlen_offsets = None
        ctx.interleaved = interleaved
        ctx.inplace = inplace
        ctx.max_seqlen = max_seqlen
        return out if not inplace else x

    @staticmethod
    def backward(ctx, do):
        seqlen_offsets = ctx.seqlen_offsets
        if seqlen_offsets is None:
            cos, sin, cu_seqlens, seqlen_offsets = ctx.saved_tensors
        else:
            cos, sin, cu_seqlens = ctx.saved_tensors
        dx = apply_rotary(
            do,
            cos,
            sin,
            seqlen_offsets=seqlen_offsets,
            cu_seqlens=cu_seqlens,
            max_seqlen=ctx.max_seqlen,
            interleaved=ctx.interleaved,
            inplace=ctx.inplace,
            conjugate=True,
        )
        return dx, None, None, None, None, None, None, None


def apply_rotary_emb(
    x,
    cos,
    sin,
    interleaved=False,
    inplace=False,
    seqlen_offsets: Union[int, Tensor] = 0,
    cu_seqlens: Optional[Tensor] = None,
    max_seqlen: Optional[int] = None,
):
    """
    Arguments:
        x: (batch_size, seqlen, nheads, headdim) if cu_seqlens is None
            else (total_seqlen, nheads, headdim)
        cos, sin: (seqlen_rotary, rotary_dim / 2)
        interleaved: if True, rotate pairs of even and odd dimensions (GPT-J style) instead
            of 1st half and 2nd half (GPT-NeoX style).
        inplace: if True, apply rotary embedding in-place.
        seqlen_offsets: (batch_size,) or int. Each sequence in x is shifted by this amount.
            Most commonly used in inference when we have KV cache.
        cu_seqlens: (batch + 1,) or None
        max_seqlen: int
    Return:
        out: (batch_size, seqlen, nheads, headdim) if cu_seqlens is None
            else (total_seqlen, nheads, headdim)
    rotary_dim must be <= headdim
    Apply rotary embedding to the first rotary_dim of x.
    """
    return ApplyRotaryEmb.apply(
        x, cos, sin, interleaved, inplace, seqlen_offsets, cu_seqlens, max_seqlen
    )


# For backward compatibility
apply_rotary_emb_func = apply_rotary_emb


def _apply_rotary_emb_qkv(
    qkv,
    cos,
    sin,
    cos_k=None,
    sin_k=None,
    interleaved=False,
    inplace=False,
    conjugate=False,
    seqlen_offsets: Union[int, Tensor] = 0,
    num_heads_q: Optional[int] = None,
):
    apply_rotary_fn = partial(
        apply_rotary,
        interleaved=interleaved,
        inplace=inplace,
        conjugate=conjugate,
        seqlen_offsets=seqlen_offsets,
    )
    if cos_k is None and sin_k is None and qkv.is_contiguous():
        # Call 1 kernel instead of 2 kernels
        # We need qkv to be contiguous so that when we reshape to combine (3, nheads)
        # dimensions, we get the same tensor
        if qkv.dim() == 5:
            batch, seqlen, three, nheads, headdim = qkv.shape
            assert three == 3
            # qk = rearrange(qkv[:, :, :2], "b s t h d -> b s (t h) d")
            qk = qkv[:, :, :2].reshape(batch, seqlen, -1, headdim)
            qk = apply_rotary_fn(qk, cos, sin)
        else:
            assert qkv.dim() == 4
            assert num_heads_q is not None
            num_heads_k = (qkv.shape[2] - num_heads_q) // 2
            assert qkv.shape[2] == num_heads_q + 2 * num_heads_k
            qk = qkv[:, :, : num_heads_q + num_heads_k]
            qk = apply_rotary_fn(qk, cos, sin)
        if not inplace:
            if qkv.dim() == 5:
                qkv = torch.cat(
                    [rearrange(qk, "b s (t h) d -> b s t h d", t=2), qkv[:, :, 2:]],
                    dim=2,
                )
            else:
                qkv = torch.cat([qk, qkv[:, :, num_heads_q + num_heads_k :]], dim=2)
    else:
        cos_k = cos if cos_k is None else cos_k
        sin_k = sin if sin_k is None else sin_k
        if qkv.dim() == 5:
            batch, seqlen, three, nheads, headdim = qkv.shape
            assert three == 3
            q, k = qkv[:, :, 0], qkv[:, :, 1]
        else:
            assert qkv.dim() == 4
            assert num_heads_q is not None
            num_heads_k = (qkv.shape[2] - num_heads_q) // 2
            assert qkv.shape[2] == num_heads_q + 2 * num_heads_k
            q, k = (
                qkv[:, :, :num_heads_q],
                qkv[:, :, num_heads_q : num_heads_q + num_heads_k],
            )
        q = apply_rotary_fn(q, cos, sin)
        k = apply_rotary_fn(k, cos_k, sin_k)
        if not inplace:
            if qkv.dim() == 5:
                qkv = torch.stack([q, k, qkv[:, :, 2]], dim=2)
            else:
                qkv = torch.cat([q, k, qkv[:, :, num_heads_q + num_heads_k :]], dim=2)
    return qkv


class ApplyRotaryEmbQKV_(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        qkv,
        cos,
        sin,
        cos_k=None,
        sin_k=None,
        interleaved=False,
        seqlen_offsets: Union[int, torch.Tensor] = 0,
        num_heads_q: Optional[int] = None,
    ):
        # apply_rotary_emb_qkv_inplace(
        qkv = _apply_rotary_emb_qkv(
            qkv,
            cos,
            sin,
            cos_k,
            sin_k,
            interleaved=interleaved,
            inplace=True,
            seqlen_offsets=seqlen_offsets,
            num_heads_q=num_heads_q,
        )
        if isinstance(seqlen_offsets, int):
            ctx.save_for_backward(cos, sin, cos_k, sin_k)
            ctx.seqlen_offsets = seqlen_offsets
        else:
            ctx.save_for_backward(cos, sin, cos_k, sin_k, seqlen_offsets)
            ctx.seqlen_offsets = None
        ctx.interleaved = interleaved
        ctx.num_heads_q = num_heads_q
        return qkv

    @staticmethod
    def backward(ctx, dqkv):
        seqlen_offsets = ctx.seqlen_offsets
        if seqlen_offsets is None:
            cos, sin, cos_k, sin_k, seqlen_offsets = ctx.saved_tensors
        else:
            cos, sin, cos_k, sin_k = ctx.saved_tensors
        dqkv = _apply_rotary_emb_qkv(
            dqkv,
            cos,
            sin,
            cos_k,
            sin_k,
            interleaved=ctx.interleaved,
            inplace=True,
            seqlen_offsets=seqlen_offsets,
            num_heads_q=ctx.num_heads_q,
            conjugate=True,
        )
        return dqkv, None, None, None, None, None, None, None


def apply_rotary_emb_qkv_(
    qkv,
    cos,
    sin,
    cos_k=None,
    sin_k=None,
    interleaved=False,
    seqlen_offsets: Union[int, torch.Tensor] = 0,
    num_heads_q: Optional[int] = None,
):
    """
    Arguments:
        qkv: (batch_size, seqlen, 3, nheads, headdim) or (batch_size, seqlen, num_heads_q + 2 * num_heads_k, headdim).
            If qkv has shape (batch_size, seqlen, num_heads_q + 2 * num_heads_k, headdim) (e.g. MQA / GQA),
            then num_heads_q must be provided.
        cos, sin: (seqlen, rotary_dim / 2)
        cos_k, sin_k: (seqlen, rotary_dim / 2), optional
        interleaved: if True, rotate pairs of even and odd dimensions (GPT-J style) instead of
            1st half and 2nd half (GPT-NeoX style).
        seqlen_offsets: (batch_size,) or int. Each sequence in Q and K is shifted by this amount.
            Most commonly used in inference when we have KV cache.
    Return:
        qkv: (batch_size, seqlen, 3, nheads, headdim) or (batch_size, seqlen, num_heads_q + 2 * num_heads_k, headdim)
    rotary_dim must be <= headdim
    Apply rotary embedding *inplace* to the first rotary_dim of Q and K.
    """
    return ApplyRotaryEmbQKV_.apply(
        qkv, cos, sin, cos_k, sin_k, interleaved, seqlen_offsets, num_heads_q
    )


class ApplyRotaryEmbKV_(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        kv,
        cos,
        sin,
        interleaved=False,
        seqlen_offsets: Union[int, torch.Tensor] = 0,
    ):
        batch, seqlen, two, nheads, headdim = kv.shape
        assert two == 2
        k = kv[:, :, 0]
        apply_rotary(
            k,
            cos,
            sin,
            seqlen_offsets=seqlen_offsets,
            interleaved=interleaved,
            inplace=True,
        )
        if isinstance(seqlen_offsets, int):
            ctx.save_for_backward(cos, sin)  # Can't save int with save_for_backward
            ctx.seqlen_offsets = seqlen_offsets
        else:
            ctx.save_for_backward(cos, sin, seqlen_offsets)
            ctx.seqlen_offsets = None
        ctx.interleaved = interleaved
        return kv

    @staticmethod
    def backward(ctx, dkv):
        seqlen_offsets = ctx.seqlen_offsets
        if seqlen_offsets is None:
            cos, sin, seqlen_offsets = ctx.saved_tensors
        else:
            cos, sin = ctx.saved_tensors
        apply_rotary(
            dkv[:, :, 0],
            cos,
            sin,
            seqlen_offsets=seqlen_offsets,
            interleaved=ctx.interleaved,
            inplace=True,
            conjugate=True,
        )
        return dkv, None, None, None, None


apply_rotary_emb_kv_ = ApplyRotaryEmbKV_.apply


def apply_rotary_emb_kv_(
    kv,
    cos,
    sin,
    interleaved=False,
    seqlen_offsets: Union[int, torch.Tensor] = 0,
):
    """
    Arguments:
        kv: (batch_size, seqlen, 2, nheads, headdim)
        cos, sin: (seqlen, rotary_dim / 2)
        interleaved: if True, rotate pairs of even and odd dimensions (GPT-J style) instead of
            1st half and 2nd half (GPT-NeoX style).
        seqlen_offsets: (batch_size,) or int. Each sequence in Q and K is shifted by this amount.
            Most commonly used in inference when we have KV cache.
    Return:
        kv: (batch_size, seqlen, 2, nheads, headdim)
    rotary_dim must be <= headdim
    Apply rotary embedding *inplace* to the first rotary_dim of K.
    """
    return ApplyRotaryEmbKV_.apply(kv, cos, sin, interleaved, seqlen_offsets)


class RotaryEmbedding(torch.nn.Module):
    """
    The rotary position embeddings from RoFormer_ (Su et. al).
    A crucial insight from the method is that the query and keys are
    transformed by rotation matrices which depend on the relative positions.

    Other implementations are available in the Rotary Transformer repo_ and in
    GPT-NeoX_, GPT-NeoX was an inspiration

    .. _RoFormer: https://arxiv.org/abs/2104.09864
    .. _repo: https://github.com/ZhuiyiTechnology/roformer
    .. _GPT-NeoX: https://github.com/EleutherAI/gpt-neox

    If scale_base is not None, this implements XPos (Sun et al., https://arxiv.org/abs/2212.10554).
    A recommended value for scale_base is 512: https://github.com/HazyResearch/flash-attention/issues/96
    Reference: https://github.com/sunyt32/torchscale/blob/main/torchscale/component/xpos_relative_position.py
    """

    def __init__(
        self,
        dim: int,
        base=10000.0,
        interleaved=False,
        scale_base=None,
        device=None,
    ):
        """
        interleaved: if True, rotate pairs of even and odd dimensions (GPT-J style) instead
            of 1st half and 2nd half (GPT-NeoX style).
        """
        super().__init__()
        self.dim = dim
        self.base = float(base)
        # Generate and save the inverse frequency buffer (non trainable)
        inv_freq = self._compute_inv_freq(device)
        self.register_buffer("inv_freq", inv_freq, persistent=False)
        self.interleaved = interleaved
        self.scale_base = scale_base
        scale = (
            (torch.arange(0, dim, 2, device=device, dtype=torch.float32) + 0.4 * dim)
            / (1.4 * dim)
            if scale_base is not None
            else None
        )
        self.register_buffer("scale", scale, persistent=False)

        self._seq_len_cached = 0
        self._cos_cached = None
        self._sin_cached = None
        self._cos_k_cached = None
        self._sin_k_cached = None

    def _compute_inv_freq(self, device=None):
        return 1.0 / (
            self.base
            ** (
                torch.arange(0, self.dim, 2, device=device, dtype=torch.float32)
                / self.dim
            )
        )

    def _update_cos_sin_cache(self, seqlen, device=None, dtype=None):
        # Reset the tables if the sequence length has changed,
        # if we're on a new device (possibly due to tracing for instance),
        # or if we're switching from inference mode to training
        if (
            seqlen > self._seq_len_cached
            or self._cos_cached is None
            or self._cos_cached.device != device
            or self._cos_cached.dtype != dtype
            or (self.training and self._cos_cached.is_inference())
        ):
            self._seq_len_cached = seqlen
            # We want fp32 here, not self.inv_freq.dtype, since the model could be loaded in bf16
            # And the output of arange can be quite large, so bf16 would lose a lot of precision.
            t = torch.arange(seqlen, device=device, dtype=torch.float32)
            # We want fp32 here as well since inv_freq will be multiplied with t, and the output
            # will be large. Having it in bf16 will lose a lot of precision and cause the
            # cos & sin output to change significantly.
            # We want to recompute self.inv_freq if it was not loaded in fp32
            if self.inv_freq.dtype != torch.float32:
                inv_freq = self._compute_inv_freq(device=device)
            else:
                inv_freq = self.inv_freq
            # Don't do einsum, it converts fp32 to bf16 under AMP
            # freqs = torch.einsum("i,j->ij", t, self.inv_freq)
            freqs = torch.outer(t, inv_freq)
            if self.scale is None:
                self._cos_cached = torch.cos(freqs).to(dtype)
                self._sin_cached = torch.sin(freqs).to(dtype)
            else:
                power = (
                    torch.arange(
                        seqlen, dtype=self.scale.dtype, device=self.scale.device
                    )
                    - seqlen // 2
                ) / self.scale_base
                scale = self.scale.to(device=power.device) ** rearrange(
                    power, "s -> s 1"
                )
                # We want the multiplication by scale to happen in fp32
                self._cos_cached = (torch.cos(freqs) * scale).to(dtype)
                self._sin_cached = (torch.sin(freqs) * scale).to(dtype)
                self._cos_k_cached = (torch.cos(freqs) / scale).to(dtype)
                self._sin_k_cached = (torch.sin(freqs) / scale).to(dtype)

    def forward(
        self,
        qkv: torch.Tensor,
        kv: Optional[torch.Tensor] = None,
        seqlen_offset: Union[int, torch.Tensor] = 0,
        max_seqlen: Optional[int] = None,
        num_heads_q: Optional[int] = None,
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        qkv: (batch, seqlen, 3, nheads, headdim) or (batch, seqlen, num_heads_q + 2 * num_heads_k, headdim)
            if kv is none, else it's just q of shape (batch, seqlen, nheads, headdim).
            If qkv has shape (batch, seqlen, num_heads_q + 2 * num_heads_k, headdim) (e.g. MQA / GQA),
            then num_heads_q must be provided.
        kv: (batch, seqlen, 2, nheads, headdim)
        seqlen_offset: (batch_size,) or int. Each sequence in x is shifted by this amount.
            Most commonly used in inference when we have KV cache.
            If it's a tensor of shape (batch_size,), then to update the cos / sin cache, one
            should pass in max_seqlen, which will update the cos / sin cache up to that length.
        Apply rotary embedding *inplace* to qkv and / or kv.
        """
        seqlen = qkv.shape[1]
        if max_seqlen is not None:
            self._update_cos_sin_cache(max_seqlen, device=qkv.device, dtype=qkv.dtype)
        elif isinstance(seqlen_offset, int):
            self._update_cos_sin_cache(
                seqlen + seqlen_offset, device=qkv.device, dtype=qkv.dtype
            )
        if kv is None:
            return apply_rotary_emb_qkv_(
                qkv,
                self._cos_cached,
                self._sin_cached,
                self._cos_k_cached if self.scale is not None else None,
                self._sin_k_cached if self.scale is not None else None,
                interleaved=self.interleaved,
                seqlen_offsets=seqlen_offset,
                num_heads_q=num_heads_q,
            )
        else:
            q = qkv
            q = apply_rotary_emb_func(
                q,
                self._cos_cached,
                self._sin_cached,
                interleaved=self.interleaved,
                inplace=True,
                seqlen_offsets=seqlen_offset,
            )
            kv = apply_rotary_emb_kv_(
                kv,
                self._cos_cached if self.scale is None else self._cos_k_cached,
                self._sin_cached if self.scale is None else self._sin_k_cached,
                interleaved=self.interleaved,
                seqlen_offsets=seqlen_offset,
            )
            return q, kv


@triton.jit
def rotary_kernel(
    OUT,  # Pointers to matrices
    X,
    COS,
    SIN,
    CU_SEQLENS,
    SEQLEN_OFFSETS,  # this could be int or a pointer
    # Matrix dimensions
    seqlen,
    nheads,
    seqlen_ro,
    # strides
    stride_out_batch,
    stride_out_seqlen,
    stride_out_nheads,
    stride_out_headdim,
    stride_x_batch,
    stride_x_seqlen,
    stride_x_nheads,
    stride_x_headdim,
    # Meta-parameters
    # We want ROTARY_DIM to be constexpr, otherwise the triton compiler doesn't know that
    # the mask is constant every 8 elements, and it will generate LDG.16 instead of LDG.128
    ROTARY_DIM: tl.constexpr,
    IS_SEQLEN_OFFSETS_TENSOR: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    INTERLEAVED: tl.constexpr,
    CONJUGATE: tl.constexpr,
    BLOCK_H: tl.constexpr,
    BLOCK_M: tl.constexpr,
):
    BLOCK_K: tl.constexpr = triton.next_power_of_2(ROTARY_DIM)
    ROTARY_DIM_HALF = ROTARY_DIM // 2
    pid_head = tl.program_id(axis=0)
    pid_m = tl.program_id(axis=1)
    pid_batch = tl.program_id(axis=2)

    if not IS_VARLEN:
        X = X + pid_batch * stride_x_batch
        OUT = OUT + pid_batch * stride_out_batch
    else:
        start_idx = tl.load(CU_SEQLENS + pid_batch)
        seqlen = tl.load(CU_SEQLENS + pid_batch + 1) - start_idx
        X = X + start_idx * stride_x_seqlen
        OUT = OUT + start_idx * stride_out_seqlen

    if pid_m * BLOCK_M >= seqlen:
        return

    rh = pid_head * BLOCK_H + tl.arange(0, BLOCK_H)
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    if not IS_SEQLEN_OFFSETS_TENSOR:
        rm_cs = rm + SEQLEN_OFFSETS
    else:
        rm_cs = rm + tl.load(SEQLEN_OFFSETS + pid_batch)

    rk_half = tl.arange(0, BLOCK_K // 2)
    COS = COS + (rm_cs[:, None] * ROTARY_DIM_HALF + rk_half[None, :])
    SIN = SIN + (rm_cs[:, None] * ROTARY_DIM_HALF + rk_half[None, :])
    mask_cs = (rm_cs[:, None] < seqlen_ro) & (rk_half[None, :] < ROTARY_DIM_HALF)
    cos = tl.load(COS, mask=mask_cs, other=1.0).to(tl.float32)
    sin = tl.load(SIN, mask=mask_cs, other=0.0).to(tl.float32)
    if CONJUGATE:
        sin = -sin

    if not INTERLEAVED:
        # Load the 1st and 2nd halves of X, do calculation, then store to 1st and 2nd halves of OUT
        X = X + (
            rh[:, None, None] * stride_x_nheads
            + rm[None, :, None] * stride_x_seqlen
            + rk_half[None, None, :] * stride_x_headdim
        )
        OUT = OUT + (
            rh[:, None, None] * stride_out_nheads
            + rm[None, :, None] * stride_out_seqlen
            + rk_half[None, None, :] * stride_out_headdim
        )
        mask = (
            (rh[:, None, None] < nheads)
            & (rm[None, :, None] < seqlen)
            & (rk_half[None, None, :] < ROTARY_DIM_HALF)
        )
        x0 = tl.load(X, mask=mask, other=0.0).to(tl.float32)
        x1 = tl.load(
            X + ROTARY_DIM_HALF * stride_x_headdim,
            mask=mask,
            other=0.0,
        ).to(tl.float32)
        o0 = x0 * cos - x1 * sin
        o1 = x0 * sin + x1 * cos
        tl.store(OUT, o0, mask=mask)
        tl.store(OUT + ROTARY_DIM_HALF * stride_out_headdim, o1, mask=mask)
    else:
        rk = tl.arange(0, BLOCK_K)
        X = X + (
            rh[:, None, None] * stride_x_nheads
            + rm[None, :, None] * stride_x_seqlen
            + rk[None, None, :] * stride_x_headdim
        )
        OUT = OUT + (
            rh[:, None, None] * stride_out_nheads
            + rm[None, :, None] * stride_out_seqlen
            + rk[None, None, :] * stride_out_headdim
        )
        mask = (
            (rh[:, None, None] < nheads)
            & (rm[None, :, None] < seqlen)
            & (rk[None, None, :] < ROTARY_DIM)
        )
        x = tl.load(X, mask=mask, other=0.0).to(tl.float32)
        x0, x1 = tl.split(tl.reshape(x, [BLOCK_H, BLOCK_M, BLOCK_K // 2, 2]))
        o0 = x0 * cos - x1 * sin
        o1 = x0 * sin + x1 * cos
        o = tl.reshape(tl.join(o0, o1), [BLOCK_H, BLOCK_M, BLOCK_K])
        tl.store(OUT, o, mask=mask)


def apply_rotary(
    x: torch.Tensor,
    cos: torch.Tensor,
    sin: torch.Tensor,
    seqlen_offsets: Union[int, torch.Tensor] = 0,
    cu_seqlens: Optional[torch.Tensor] = None,
    max_seqlen: Optional[int] = None,
    interleaved=False,
    inplace=False,
    conjugate=False,
) -> torch.Tensor:
    """
    Arguments:
        x: (batch, seqlen, nheads, headdim) if cu_seqlens is None
            else (total_seqlen, nheads, headdim).
        cos: (seqlen_ro, rotary_dim / 2)
        sin: (seqlen_ro, rotary_dim / 2)
        seqlen_offsets: integer or integer tensor of size (batch,)
        cu_seqlens: (batch + 1,) or None
        max_seqlen: int
    Returns:
        y: (batch, seqlen, nheads, headdim)
    """
    is_varlen = cu_seqlens is not None
    if not is_varlen:
        batch, seqlen, nheads, headdim = x.shape
    else:
        assert max_seqlen is not None, (
            "If cu_seqlens is passed in, then max_seqlen must be passed"
        )
        total_seqlen, nheads, headdim = x.shape
        batch_p_1 = cu_seqlens.shape[0]
        batch = batch_p_1 - 1
        seqlen = max_seqlen
    seqlen_ro, rotary_dim = cos.shape
    assert sin.shape == cos.shape
    rotary_dim *= 2
    assert rotary_dim <= headdim, "rotary_dim must be <= headdim"
    assert headdim <= 256, "Only support headdim <= 256"
    assert seqlen_ro >= seqlen, "seqlen_ro must be >= seqlen"

    cos, sin = cos.contiguous(), sin.contiguous()
    if isinstance(seqlen_offsets, torch.Tensor):
        assert seqlen_offsets.shape == (batch,)
        assert seqlen_offsets.dtype in [torch.int32, torch.int64]
        seqlen_offsets = seqlen_offsets.contiguous()
    else:
        assert seqlen_offsets + seqlen <= seqlen_ro

    output = torch.empty_like(x) if not inplace else x
    if rotary_dim < headdim and not inplace:
        output[..., rotary_dim:].copy_(x[..., rotary_dim:])

    grid = lambda META: (
        triton.cdiv(nheads, META["BLOCK_H"]),
        triton.cdiv(seqlen, META["BLOCK_M"]),
        batch,
    )  # noqa
    BLOCK_M = 8 if rotary_dim <= 128 else 4

    # Need this, otherwise Triton tries to launch from cuda:0 and we get
    # ValueError: Pointer argument (at 0) cannot be accessed from Triton (cpu tensor?)
    with torch.cuda.device(x.device.index):
        torch.library.wrap_triton(rotary_kernel)[grid](
            output,  # data ptrs
            x,
            cos,
            sin,
            cu_seqlens,
            seqlen_offsets,
            seqlen,  # shapes
            nheads,
            seqlen_ro,
            output.stride(0)
            if not is_varlen
            else 0,  # batch_strides if not varlen else 0
            output.stride(-3),  # seqlen_stride or total_seqlen_stride
            output.stride(-2),  # nheads_stride
            output.stride(-1),  # headdim_stride
            x.stride(0) if not is_varlen else 0,  # batch_strides if not varlen else 0
            x.stride(-3),  # seqlen stride or total_seqlen_stride
            x.stride(-2),  # nheads stride
            x.stride(-1),  # headdim stride
            rotary_dim,
            isinstance(seqlen_offsets, torch.Tensor),
            is_varlen,
            interleaved,
            conjugate,
            BLOCK_M=BLOCK_M,
            BLOCK_H=2,
        )
    return output



import torch
from sageattention import sageattn
from torch.nn.attention import SDPBackend, sdpa_kernel
from torch.nn.functional import scaled_dot_product_attention


def attention(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    dtype: torch.dtype = torch.bfloat16,
):
    """Attention function
    Attention(o) = softmax(Q @K^T) @ V
    q: [B, L, N, D]
    k: [B, L, N, D]
    v: [B, L, N, D]
    dtype: torch.dtype. Apply when dtype of q/k/v is not float16/bfloat16.
    """
    # print(f"attention: q shape: {q.shape}, k shape: {k.shape}, v shape: {v.shape}")
    half_dtypes = (torch.float16, torch.bfloat16)
    assert dtype in half_dtypes
    assert q.device.type == "cuda" and q.size(-1) <= 256
    _B, L_k, _N_k, _D = k.shape

    def half(x):
        return x if x.dtype in half_dtypes else x.to(dtype)

    # Convert to half precision if needed
    q = half(q)
    k = half(k)
    v = half(v)

    q = q.to(v.dtype)
    k = k.to(v.dtype)
    if L_k < 512:
        with sdpa_kernel(SDPBackend.FLASH_ATTENTION):
            return scaled_dot_product_attention(
                q.permute(0, 2, 1, 3), k.permute(0, 2, 1, 3), v.permute(0, 2, 1, 3)
            ).permute(0, 2, 1, 3)
    else:
        return sageattn(q=q, k=k, v=v, tensor_layout="NHD", output_dtype=dtype)



import time
from typing import List

import torch
from loguru import logger

from .gemlite.helper import (
    patch_model,
    A8W8_int8_dynamic,
)


def quantize_model_a8w8_int8_gemlite(
    model: torch.nn.Module,
    device: str = "cuda",
    exclude: List[str] = [
        "time_embedding.0",
        "time_embedding.2",
        "time_projection.1",
        "head.head",
        "img_emb.proj.1",
        "img_emb.proj.3",
    ],
) -> None:

    logger.info(f"Quantizing model on {device}...")
    quant_start_time = time.perf_counter()
    patch_model(
        model, processor=A8W8_int8_dynamic(), device=device, skip_modules=exclude
    )
    quant_end_time = time.perf_counter()
    logger.info(f"Quantization time: {quant_end_time - quant_start_time:.2f} seconds")



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import torch
import copy
from torch import Tensor
import triton
import triton.language as tl
from ..dtypes import is_mx_dtype
from .config import AUTOTUNE
from .utils import *

KEYS = ["M", "N", "K", "group_size", "elements_per_sample", "type_id"]
MATMUL_TYPE = "GEMV_SPLITK"
NATIVE_ATOMIC = gpu_supports_bfloat16_atomicadd()


def kernel_config_pruner(configs, nargs, **kwargs):
    global KEYS
    from ..core import GEMLITE_TRITON_CONFIG_CACHE

    m = nargs["M"]
    n = nargs["N"]
    k = nargs["K"]
    g = nargs["group_size"]
    e = nargs["elements_per_sample"]

    # Check cache
    if MATMUL_TYPE in GEMLITE_TRITON_CONFIG_CACHE:
        signature = str(tuple([nargs[i] for i in KEYS]))
        if signature in GEMLITE_TRITON_CONFIG_CACHE[MATMUL_TYPE]:
            config = copy.deepcopy(GEMLITE_TRITON_CONFIG_CACHE[MATMUL_TYPE][signature])
            num_stages = config.pop("num_stages")
            num_warps = config.pop("num_warps")
            num_ctas = config.pop("num_ctas")

            config.pop("num_buffers_warp_spec", None)
            config.pop("num_consumer_groups", None)
            config.pop("reg_dec_producer", None)
            config.pop("reg_inc_consumer", None)

            yield triton.Config(
                config,
                num_stages=num_stages,
                num_warps=num_warps,
                pre_hook=init_to_zero("c_ptr") if config["SPLIT_K"] > 1 else None,
            )

            return

    used = set()
    for config in configs:
        group_size_m = config.kwargs["GROUP_SIZE_M"]
        block_size_m = 1  # Only 1 is supported
        block_size_n = min(n, config.kwargs["BLOCK_SIZE_N"])
        block_size_k = min(k, config.kwargs["BLOCK_SIZE_K"])
        split_k = config.kwargs["SPLIT_K"]

        num_stages = config.num_stages
        num_warps = config.num_warps
        A_load_order = config.kwargs["A_load_order"]
        dot_prod_mode = config.kwargs["dot_prod_mode"]

        ##Constraints: BLOCK_SIZE_K <= group_size -> load_scales_as_block is always False for gemvs
        block_size_k = min(block_size_k, g)
        block_size_k = next_power_of_2(block_size_k)
        block_size_n = next_power_of_2(block_size_n)

        # K needs to be divisible by BLOCK_SIZE_K * SPLIT_K: TODO: without this, cuda-graphs breaks.
        while block_size_k > 16 and not is_divisible(k, block_size_k * split_k):
            block_size_k //= 2

        # Skip blocks that are either too large or too small
        block_area = (block_size_k // split_k) * block_size_n
        if block_area < 1024 or block_area > 4096 * 8:  # 128 * 8 * num_warps
            continue

        # Block size should be compatible with minimum-packing
        if block_size_k < e:
            continue

        key = (
            block_size_m,
            block_size_n,
            block_size_k,
            group_size_m,
            split_k,
            A_load_order,
            dot_prod_mode,
            num_stages,
            num_warps,
        )

        new_config = {
            "BLOCK_SIZE_M": block_size_m,
            "BLOCK_SIZE_N": block_size_n,
            "BLOCK_SIZE_K": block_size_k,
            "GROUP_SIZE_M": group_size_m,
            "SPLIT_K": split_k,
            "A_load_order": A_load_order,
            "dot_prod_mode": dot_prod_mode,
        }

        if IS_HIP:
            new_config["waves_per_eu"] = config.kwargs.get("waves_per_eu", 0)
            key = key + (new_config["waves_per_eu"],)

        if key in used:
            continue

        used.add(key)
        yield triton.Config(
            new_config,
            num_stages=num_stages,
            num_warps=num_warps,
            pre_hook=init_to_zero("c_ptr") if split_k > 1 else None,
        )


########################################################################################################################################################################
# Nvidia


# contiguous=False
def get_max_autotune_config_nvidia():
    _configs = []
    for A in [0, 1]:  # [0, 1, 2, 3]
        for D in [0]:  # 1: breaks with fp8
            for w in [4, 8]:
                for s in [1, 2]:
                    for M in [1]:
                        for N in [1, 2, 4, 8, 16, 32, 64]:
                            for K in [32, 64, 128, 256, 512, 1024, 2048, 4096]:
                                for SK in [1]:
                                    _configs.append(
                                        triton.Config(
                                            {
                                                "BLOCK_SIZE_M": M,
                                                "BLOCK_SIZE_N": N,
                                                "BLOCK_SIZE_K": K,
                                                "GROUP_SIZE_M": 8,
                                                "SPLIT_K": SK,
                                                "A_load_order": A,
                                                "dot_prod_mode": D,
                                            },
                                            num_stages=s,
                                            num_warps=w,
                                        )
                                    )

    return _configs


def get_fast_autotune_config_nvidia():
    configs = []

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 16,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=4,
            num_stages=2,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 2,
                "BLOCK_SIZE_K": 512,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=8,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 512,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 1,
                "dot_prod_mode": 0,
            },
            num_warps=4,
            num_stages=1,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 1,
                "BLOCK_SIZE_K": 1024,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 1,
                "dot_prod_mode": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 1,
                "BLOCK_SIZE_K": 1024,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=4,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 1024,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 8,
                "BLOCK_SIZE_K": 1024,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=4,
            num_stages=2,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 1,
                "BLOCK_SIZE_K": 2048,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 2,
                "BLOCK_SIZE_K": 2048,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=4,
            num_stages=1,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 2,
                "BLOCK_SIZE_K": 4096,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=8,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 4096,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
            },
            num_warps=4,
            num_stages=1,
        )
    )

    return configs


def get_default_config_nvidia():
    config = triton.Config(
        {
            "BLOCK_SIZE_M": 1,
            "BLOCK_SIZE_N": 2,
            "BLOCK_SIZE_K": 2048,
            "GROUP_SIZE_M": 8,
            "SPLIT_K": 1,
            "A_load_order": 1,
            "dot_prod_mode": 0,
        },
        num_warps=4,
        num_stages=2,
    )

    return [config]


########################################################################################################################################################################
# AMD - Instinct MI300X


# contiguous = False (For unpacked data)
def get_max_autotune_config_amd():
    _configs = []
    for A in [0]:
        for D in [0]:  # 1: breaks with fp8
            for w in [4]:
                for s in [1, 2]:
                    for v in [0, 2, 4]:
                        for M in [1]:
                            for N in [1, 2, 4, 8, 16, 32, 64]:
                                for K in [32, 64, 128, 256, 512, 1024, 2048, 4096]:
                                    for SK in [1]:
                                        _configs.append(
                                            triton.Config(
                                                {
                                                    "BLOCK_SIZE_M": M,
                                                    "BLOCK_SIZE_N": N,
                                                    "BLOCK_SIZE_K": K,
                                                    "GROUP_SIZE_M": 8,
                                                    "SPLIT_K": SK,
                                                    "A_load_order": A,
                                                    "dot_prod_mode": D,
                                                    "waves_per_eu": v,
                                                },
                                                num_stages=s,
                                                num_warps=w,
                                            )
                                        )

    return _configs


def get_fast_autotune_config_amd():
    configs = []

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 16,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 0,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 16,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 2,
                "BLOCK_SIZE_K": 512,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 0,
            },
            num_warps=4,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 512,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 8,
                "BLOCK_SIZE_K": 512,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 4,
            },
            num_warps=4,
            num_stages=1,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 1,
                "BLOCK_SIZE_K": 1024,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 1024,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 4,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 8,
                "BLOCK_SIZE_K": 1024,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 4,
            },
            num_warps=4,
            num_stages=1,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 1,
                "BLOCK_SIZE_K": 2048,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 2,
                "BLOCK_SIZE_K": 2048,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 8,
                "BLOCK_SIZE_K": 2048,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 0,
            },
            num_warps=4,
            num_stages=1,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 1,
                "BLOCK_SIZE_K": 4096,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 0,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 2,
                "BLOCK_SIZE_K": 4096,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 1,
                "BLOCK_SIZE_N": 4,
                "BLOCK_SIZE_K": 4096,
                "GROUP_SIZE_M": 8,
                "SPLIT_K": 1,
                "A_load_order": 0,
                "dot_prod_mode": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=1,
        )
    )

    return configs


def get_default_config_amd():
    config = triton.Config(
        {
            "BLOCK_SIZE_M": 1,
            "BLOCK_SIZE_N": 2,
            "BLOCK_SIZE_K": 2048,
            "GROUP_SIZE_M": 8,
            "SPLIT_K": 1,
            "A_load_order": 1,
            "dot_prod_mode": 0,
        },
        num_warps=2,
        num_stages=1,
    )

    return [config]


########################################################################################################################################################################

if IS_HIP:
    get_max_autotune_config = get_max_autotune_config_amd
    get_fast_autotune_config = get_fast_autotune_config_amd
    get_default_config = get_default_config_amd
else:
    get_max_autotune_config = get_max_autotune_config_nvidia
    get_fast_autotune_config = get_fast_autotune_config_nvidia
    get_default_config = get_default_config_nvidia

AUTOTUNE_SETTING = AUTOTUNE.GEMV_SPLITK
if AUTOTUNE_SETTING == "max":
    get_autotune_config = get_max_autotune_config
elif AUTOTUNE_SETTING == "fast":
    get_autotune_config = get_fast_autotune_config
else:
    get_autotune_config = get_default_config


@triton.autotune(
    configs=get_autotune_config(),
    key=KEYS,
    prune_configs_by={"early_config_prune": kernel_config_pruner},
    use_cuda_graph=AUTOTUNE.USE_CUDA_GRAPH,
)
@triton.jit
def gemv_INT_splitK_kernel(
    a_ptr,
    b_ptr,
    c_ptr,
    scales_ptr,
    zeros_ptr,
    scales_a_ptr,
    M,
    N,
    K,
    ######### Quant parms #########
    W_nbits: tl.constexpr,
    group_size: tl.constexpr,
    unpack_mask: tl.constexpr,
    elements_per_sample: tl.constexpr,
    type_id: tl.constexpr,
    ######### Strides #########
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    stride_meta_g,
    stride_meta_n,
    ######### Dtypes #########
    input_dtype: tl.constexpr,
    output_dtype: tl.constexpr,
    acc_dtype: tl.constexpr,
    meta_dtype: tl.constexpr,
    ######### Meta-data mode #########
    channel_scale_mode: tl.constexpr,
    W_group_mode: tl.constexpr,
    zero_is_scalar: tl.constexpr,
    ######### tuning params #########
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    SPLIT_K: tl.constexpr,
    A_load_order: tl.constexpr,
    dot_prod_mode: tl.constexpr,
    data_contiguous: tl.constexpr,
    dump_b_val: tl.constexpr = 0,  # Improve accuracy mainly for A16W8 with post looop scaling
    #################################
    meta_evict_policy: tl.constexpr = "",
    atomic_mode: tl.constexpr = "relaxed",
    a_evict: tl.constexpr = "evict_last",
    b_evict: tl.constexpr = "evict_first",
):
    """
    Based on https://github.com/foundation-model-stack/foundation-model-stack/blob/triton/triton/kernels/gptq/splitk_dequant_gemm.py
    GEMM for C = matmul(A, dequantize(B, scales, zeros))
    A is of shape (M, K): float16 or bfloat16
    B is of shape (K//elements_per_sample, N): int32 as a packed matrix
    C is of shape (M, N): float16 or bfloat16 depending on the input A
    scales and zeros is of shape (group_size, N): float16 or bfloat16

    BLOCK_SIZE_M >=16
    BLOCK_SIZE_K * SPLIT_K <= group_size for imp1
    BLOCK_SIZE_K == SPLIT_K for imp2 (similar to original)
    """

    pid = tl.program_id(axis=0)
    pid_k = tl.program_id(axis=1)

    # Swizzle?
    pid_m, pid_n = linear_tile(pid, M, N, BLOCK_SIZE_M, BLOCK_SIZE_N, None)
    # pid_m, pid_n = swizzle_tile(pid, M, N, BLOCK_SIZE_M, BLOCK_SIZE_N, GROUP_SIZE_M)

    num_pid_k = tl.cdiv(K, BLOCK_SIZE_K * SPLIT_K)

    # Offsets
    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_k = pid_k * BLOCK_SIZE_K + tl.arange(0, BLOCK_SIZE_K)

    # Vectorized coalesced load
    ##############################
    if data_contiguous:
        offs_bn = offs_n
    else:
        offs_bn = tl.max_contiguous(tl.multiple_of(offs_n, BLOCK_SIZE_N), BLOCK_SIZE_N)
    offs_am = tl.max_contiguous(tl.multiple_of(offs_m, BLOCK_SIZE_M), BLOCK_SIZE_M)
    offs_ak = offs_k
    offs_bk = offs_k
    ###############################

    # Inputs
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_ak[None, :] * stride_ak)
    a_mask = ((offs_am[:, None] < M) & (offs_ak[None, :] < K)).to(tl.int1)
    b_ptrs = b_ptr + (
        (offs_bk[:, None] // elements_per_sample) * stride_bk
        + offs_bn[None, :] * stride_bn
    )

    # Meta data stuff
    q_shift = ((offs_k % elements_per_sample) * W_nbits).to(tl.int32)[:, None]

    scales_ptrs = scales_ptr + offs_bn[None, :] * stride_meta_n
    zeros_ptrs = zeros_ptr + offs_bn[None, :] * stride_meta_n
    stride_mul = BLOCK_SIZE_K / group_size

    BLOCK_SIZE_K_U: tl.constexpr = (BLOCK_SIZE_K) * SPLIT_K
    BLOCK_SIZE_K_P: tl.constexpr = (BLOCK_SIZE_K // elements_per_sample) * SPLIT_K

    if zero_is_scalar:
        zero_scalar = tl.load(zeros_ptr, eviction_policy="evict_last")
    ##################################################################

    if dot_prod_mode == 0:
        acc = tl.zeros((BLOCK_SIZE_K, BLOCK_SIZE_N), dtype=acc_dtype)
    if dot_prod_mode == 1:
        acc = tl.zeros((1, BLOCK_SIZE_N), dtype=acc_dtype)

    # for k in tl.range(0, num_pid_k, 1, num_stages=1):
    for k in range(num_pid_k):
        if A_load_order == 0:  # Early load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        b = tl.load(b_ptrs, eviction_policy=b_evict)

        if A_load_order == 1:  # Early load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        if W_group_mode > 0:
            k_m = ((k * SPLIT_K + pid_k) * stride_mul).to(tl.int32)

        if W_group_mode >= 2:  # [2, 3, 4]
            scales = tl.load(
                scales_ptrs + k_m * stride_meta_g, eviction_policy=meta_evict_policy
            )
        else:
            scales = None

        if W_group_mode == 1 or W_group_mode >= 3:  # [1, 3, 4]
            if zero_is_scalar:
                zeros = zero_scalar
            else:
                zeros = tl.load(
                    zeros_ptrs + k_m * stride_meta_g, eviction_policy=meta_evict_policy
                )
        else:
            zeros = None

        if A_load_order == 2:  # Mid load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        # Unpack and dequantize
        b = dequantize(
            b,
            scales,
            zeros,
            q_shift,
            meta_dtype,
            unpack_mask,
            elements_per_sample,
            W_group_mode,
            zero_is_scalar,
        )

        if A_load_order == 3:  # Late load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        if dump_b_val > 0:
            b = b.to(tl.float32) * dump_b_val

        if dot_prod_mode == 0:
            acc += a.reshape((BLOCK_SIZE_K, 1), can_reorder=False).to(acc_dtype) * b.to(
                acc_dtype
            )
        if dot_prod_mode == 1:
            acc += tl.sum(
                a.reshape((BLOCK_SIZE_K, 1), can_reorder=False) * b.to(input_dtype),
                axis=0,
                keep_dims=True,
            ).to(acc_dtype)

        # Advance
        a_ptrs += BLOCK_SIZE_K_U * stride_ak
        b_ptrs += BLOCK_SIZE_K_P * stride_bk

    if dot_prod_mode == 0:
        acc = tl.sum(acc, axis=0, keep_dims=True)

    if dump_b_val > 0:
        acc /= dump_b_val

    ##################################################################
    # Channel-wise scaling
    if channel_scale_mode == 1:  # weight-only
        scales_b = tl.load(
            scales_ptr + offs_bn,
            mask=offs_bn < N,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        acc = acc.to(meta_dtype) * scales_b[None, :]

    if channel_scale_mode == 2:  # activation-only
        scales_a = tl.load(
            scales_a_ptr + offs_am,
            mask=offs_am < M,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        scales_b = tl.full((BLOCK_SIZE_N,), value=1, dtype=meta_dtype)
        acc = acc.to(meta_dtype) * (scales_a[:, None] * scales_b[None, :])

    if channel_scale_mode == 3:  # weight + activation
        scales_a = tl.load(
            scales_a_ptr + offs_am,
            mask=offs_am < M,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        scales_b = tl.load(
            scales_ptr + offs_bn,
            mask=offs_bn < N,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        acc = acc.to(meta_dtype) * (scales_a[:, None] * scales_b[None, :])

    ##################################################################

    # Output
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_cn = tl.max_contiguous(tl.multiple_of(offs_cn, BLOCK_SIZE_N), BLOCK_SIZE_N)
    c_ptrs = c_ptr + (offs_cm[:, None] * stride_cm + offs_cn[None, :] * stride_cn)
    mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)

    if SPLIT_K == 1:
        tl.store(c_ptrs, acc, mask=mask)
    else:
        tl.atomic_add(c_ptrs, acc, mask=mask, sem=atomic_mode)


def gemv_splitK_forward(
    x: Tensor,
    W_q: Tensor,
    scales: Tensor,
    zeros: Tensor,
    scales_x: Tensor,
    W_nbits: int,
    group_size: int,
    unpack_mask: int,
    elements_per_sample: int,
    input_dtype: int,
    output_dtype: int,
    acc_dtype: int,
    meta_dtype: int,
    channel_scale_mode: int,
    W_group_mode: int,
    data_contiguous: bool,
    type_id: int,
) -> Tensor:

    M, K, N = x.shape[0], x.shape[1], W_q.shape[1]
    # assert K == W_q.shape[0] * elements_per_sample, "Invalid Input Shapes"

    native_atomic = True
    # native_atomic = (output_dtype in [DType.FP16.value, DType.FP32.value]) or NATIVE_ATOMIC
    # WARNING: change this to the second check if SPLIT_K > 1

    output = torch.empty(
        (M, N),
        device=W_q.device,
        dtype=DTYPE_TO_TORCH[output_dtype] if native_atomic else torch.float32,
    )

    grid = lambda META: (
        triton.cdiv(M, META["BLOCK_SIZE_M"]) * triton.cdiv(N, META["BLOCK_SIZE_N"]),
        META["SPLIT_K"],
    )

    dtype = DTYPE_TO_TRITON[input_dtype]
    if dtype in [tl.float16, tl.bfloat16, tl.float32]:
        acc_dtype = dtype
    else:
        acc_dtype = DTYPE_TO_TRITON[acc_dtype]

    if is_mx_dtype(input_dtype):
        raise Exception("MX dtypes are not supported for this kernel.")
    else:
        gemv_splitK_kernel = gemv_INT_splitK_kernel

    gemv_splitK_kernel[grid](
        x,
        W_q,
        output,
        scales,
        zeros,
        scales_x,
        M,
        N,
        K,
        W_nbits,
        group_size,
        unpack_mask,
        elements_per_sample,
        type_id,
        x.stride(0),
        x.stride(1),
        W_q.stride(0),
        W_q.stride(1),
        output.stride(0),
        output.stride(1),
        scales.stride(0),
        scales.stride(1),
        ################################################
        input_dtype=DTYPE_TO_TRITON[input_dtype],
        output_dtype=TORCH_DTYPE_TO_TRITON[output.dtype],
        acc_dtype=acc_dtype,
        meta_dtype=DTYPE_TO_TRITON[meta_dtype],
        ################################################
        channel_scale_mode=channel_scale_mode,
        W_group_mode=W_group_mode,
        zero_is_scalar=zeros.numel() == 1,
        data_contiguous=data_contiguous,
        dump_b_val=0.001
        if (W_group_mode in [0, 1] and acc_dtype == DType.FP16.value and W_nbits == 8)
        else 0,  # Warning: Only use with INT8
    )

    if not native_atomic:
        output = output.to(DTYPE_TO_TORCH[output_dtype])

    return output


class gemv_splitK:
    kernel = [gemv_INT_splitK_kernel]
    forward = gemv_splitK_forward
    matmul_type = MATMUL_TYPE


__all__ = ["gemv_splitK"]



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import torch
import triton
import triton.language as tl
from .dtypes import TORCH_DTYPE_TO_TRITON, PACKING_BITWIDTH_TO_TORCH_DTYPE


# Pack data, adapted from: following the same logic as: https://github.com/LeiWang1999/AutoGPTQ.tvm/blob/dcd135b9784b9f98235fc91467fe3c3c8afa34fc/auto_gptq/nn_modules/qlinear_triton.py#L413-L419
def pack_weights_over_rows_torch(
    W_q: torch.Tensor, W_nbits: int, packing_bitwidth: int, transpose: bool
) -> tuple[torch.Tensor, int]:

    assert packing_bitwidth in [8, 16, 32, 64], "Unsuported bitpacking width"
    assert W_nbits in [8, 4, 2, 1], "Unsuported nbits"
    elements_per_sample = packing_bitwidth // W_nbits

    W_q = W_q.to(torch.int32)
    W_q_out = torch.zeros(
        (W_q.shape[0] // elements_per_sample, W_q.shape[1]),
        dtype=torch.int32 if packing_bitwidth <= 32 else torch.int64,
        device=W_q.device,
    )

    for j in range(W_q.shape[0]):
        row = j // elements_per_sample
        offset = j % elements_per_sample
        W_q_out[row] |= W_q[j] << (W_nbits * offset)

    W_q_out = W_q_out.to(dtype=PACKING_BITWIDTH_TO_TORCH_DTYPE[packing_bitwidth])

    if transpose:
        W_q_out = W_q_out.t()

    return W_q_out, elements_per_sample


def pack_weights_over_cols_torch(
    W_q: torch.Tensor, W_nbits: int, packing_bitwidth: int, transpose: bool
) -> tuple[torch.Tensor, int]:

    assert packing_bitwidth in [8, 16, 32, 64], "Unsuported bitpacking width"
    assert W_nbits in [8, 4, 2, 1], "Unsuported nbits"
    elements_per_sample = packing_bitwidth // W_nbits

    W_q = W_q.to(torch.int32)
    W_q_out = torch.zeros(
        (W_q.shape[0], W_q.shape[1] // elements_per_sample),
        dtype=torch.int32 if packing_bitwidth <= 32 else torch.int64,
        device=W_q.device,
    )

    for j in range(W_q.shape[1]):
        col = j // elements_per_sample
        shift = (j % elements_per_sample) * W_nbits
        W_q_out[:, col] |= W_q[:, j] << shift

    W_q_out = W_q_out.to(dtype=PACKING_BITWIDTH_TO_TORCH_DTYPE[packing_bitwidth])

    if transpose:
        W_q_out = W_q_out.t()

    return W_q_out, elements_per_sample


############################################################################################
# Triton Bitpacking
############################################################################################
_powers_of_2 = [2**n for n in range(10)][::-1]


def highest_divisor(n: int, max_val: int) -> int:
    if max_val == 1:
        return 1

    for d in _powers_of_2:
        if n % d == 0 and d <= max_val:
            return d


@triton.jit
def or_fn(a, b):
    return a | b


@triton.jit
def pack_weights_over_cols_kernel(
    W_q_ptr,
    W_q_out_ptr,
    num_input_cols,
    num_cols,
    unroll: tl.constexpr,
    elements_per_sample: tl.constexpr,
    W_nbits: tl.constexpr,
    out_dtype: tl.constexpr,
):

    pid = tl.program_id(0)
    pid_row = (pid // num_cols) * unroll
    pid_col = pid % num_cols

    for r in range(unroll):
        start_col = pid_col * elements_per_sample
        cols = tl.arange(0, elements_per_sample)
        shifts = (cols * W_nbits).to(out_dtype)

        # Load
        offset = pid_row * num_input_cols + start_col + cols
        offset = tl.max_contiguous(
            tl.multiple_of(offset, elements_per_sample), elements_per_sample
        )
        values = tl.load(W_q_ptr + offset).to(out_dtype)

        # Pack
        result = tl.reduce(values << shifts, axis=0, combine_fn=or_fn)

        # Store
        output_offset = pid_row * num_cols + pid_col
        tl.store(W_q_out_ptr + output_offset, result)
        pid_row += 1


def pack_weights_over_cols_triton(
    W_q: torch.Tensor, W_nbits: int, packing_bitwidth: int, transpose: bool
) -> tuple[torch.Tensor, int]:

    assert packing_bitwidth in [8, 16, 32, 64], "Unsuported bitpacking width"
    assert W_nbits in [8, 4, 2, 1], "Unsuported nbits"
    elements_per_sample = packing_bitwidth // W_nbits
    num_rows, num_input_cols = W_q.shape
    num_cols = num_input_cols // elements_per_sample

    dtype = PACKING_BITWIDTH_TO_TORCH_DTYPE[packing_bitwidth]
    out_dtype = TORCH_DTYPE_TO_TRITON[dtype]

    W_q_out = torch.empty((num_rows, num_cols), dtype=dtype, device=W_q.device)
    unroll = highest_divisor(num_rows, max_val=64)
    grid = (triton.cdiv(num_rows * num_cols, unroll),)

    pack_weights_over_cols_kernel[grid](
        W_q.contiguous(),
        W_q_out,
        num_input_cols,
        num_cols,
        unroll,
        elements_per_sample,
        W_nbits,
        out_dtype,
        num_stages=2,
        num_warps=1,
    )

    if transpose:
        W_q_out = W_q_out.t()

    return W_q_out, elements_per_sample


@torch.library.custom_op("gemlite::unpack_over_cols_torch", mutates_args=())
def unpack_over_cols_torch(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_cols: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:

    num_rows, num_cols = W_q_packed.shape
    elements_per_sample = num_output_cols // num_cols

    shifts = (
        torch.arange(
            elements_per_sample, device=W_q_packed.device, dtype=W_q_packed.dtype
        )
        * W_nbits
    )
    mask = (1 << W_nbits) - 1
    W_q_unpacked = ((W_q_packed.unsqueeze(-1) >> shifts) & mask).to(dtype)
    W_q_unpacked = W_q_unpacked.view(num_rows, num_output_cols)

    return W_q_unpacked


@torch.library.register_fake("gemlite::unpack_over_cols_torch")
def unpack_over_cols_torch_fake(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_rows: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:

    num_rows, num_cols = W_q_packed.shape
    return torch.empty(
        (num_rows, num_output_cols), dtype=dtype, device=W_q_packed.device
    )


@triton.jit
def unpack_over_cols_kernel(
    W_q_packed_ptr,
    W_q_unpacked_ptr,
    num_rows,
    num_cols,
    num_output_cols,
    elements_per_sample: tl.constexpr,
    W_nbits: tl.constexpr,
    unroll: tl.constexpr,
    output_dtype: tl.constexpr,
):
    pid = tl.program_id(0)
    num_blocks = tl.cdiv(num_output_cols, unroll)
    pid_row = pid // num_blocks
    pid_col_block = pid % num_blocks

    # Load
    cols = pid_col_block * unroll + tl.arange(0, unroll)
    packed_cols = cols // elements_per_sample
    offset = pid_row * num_cols + packed_cols
    packed_values = tl.load(W_q_packed_ptr + offset)

    # Unpack
    shifts = (cols % elements_per_sample) * W_nbits
    mask_val = (1 << W_nbits) - 1
    unpacked_values = ((packed_values >> shifts) & mask_val).to(output_dtype)

    # Store the unpacked values
    unpacked_offsets = pid_row * num_output_cols + cols
    tl.store(W_q_unpacked_ptr + unpacked_offsets, unpacked_values)


@torch.library.custom_op("gemlite::unpack_over_cols_triton", mutates_args=())
def unpack_over_cols_triton(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_cols: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:

    # Get input dimensions
    num_rows, num_cols = W_q_packed.shape
    elements_per_sample = num_output_cols // num_cols

    # Allocate output tensor
    W_q_unpacked = torch.empty(
        (num_rows, num_output_cols), dtype=dtype, device=W_q_packed.device
    )
    output_dtype = TORCH_DTYPE_TO_TRITON[dtype]

    unroll = highest_divisor(num_cols, max_val=256)
    grid = (num_rows * triton.cdiv(num_output_cols, unroll),)

    # Launch the kernel
    unpack_over_cols_kernel[grid](
        W_q_packed.contiguous(),
        W_q_unpacked,
        num_rows,
        num_cols,
        num_output_cols,
        elements_per_sample,
        W_nbits,
        unroll,
        output_dtype,
        num_stages=2,
        num_warps=1,
    )

    return W_q_unpacked


@torch.library.register_fake("gemlite::unpack_over_cols_triton")
def unpack_over_cols_triton_fake(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_rows: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:
    num_rows, num_cols = W_q_packed.shape
    return torch.empty(
        (num_rows, num_output_cols), dtype=dtype, device=W_q_packed.device
    )


@triton.jit
def pack_weights_over_rows_kernel(
    W_q_ptr,
    W_q_out_ptr,
    num_rows,
    num_cols,
    unroll: tl.constexpr,
    elements_per_sample: tl.constexpr,
    W_nbits: tl.constexpr,
    out_dtype: tl.constexpr,
):

    pid = tl.program_id(0)
    num_blocks = tl.cdiv(num_cols, unroll)
    pid_row = pid // num_blocks
    pid_col = pid % num_blocks
    col = pid_col * unroll

    for r in range(unroll):
        start_row = pid_row * elements_per_sample
        rows = tl.arange(0, elements_per_sample)

        # Load
        offset = (start_row + rows) * num_cols + col
        values = tl.load(W_q_ptr + offset).to(out_dtype)

        # Pack
        shifts = (rows * W_nbits).to(out_dtype)
        result = tl.reduce(values << shifts, axis=0, combine_fn=or_fn)

        # Store
        output_offset = pid_row * num_cols + col
        tl.store(W_q_out_ptr + output_offset, result)
        col += 1


def pack_weights_over_rows_triton(
    W_q: torch.Tensor, W_nbits: int, packing_bitwidth: int, transpose: bool
) -> tuple[torch.Tensor, int]:

    elements_per_sample = packing_bitwidth // W_nbits
    num_input_rows, num_cols = W_q.shape
    num_rows = num_input_rows // elements_per_sample

    dtype = PACKING_BITWIDTH_TO_TORCH_DTYPE[packing_bitwidth]
    out_dtype = TORCH_DTYPE_TO_TRITON[dtype]

    W_q_out = torch.empty((num_rows, num_cols), dtype=dtype, device=W_q.device)
    unroll = highest_divisor(num_cols, max_val=64)
    grid = (triton.cdiv(num_rows * num_cols, unroll),)

    pack_weights_over_rows_kernel[grid](
        W_q.contiguous(),
        W_q_out,
        num_rows,
        num_cols,
        unroll,
        elements_per_sample,
        W_nbits,
        out_dtype,
        num_stages=2,
        num_warps=1,
    )

    if transpose:
        W_q_out = W_q_out.t()

    return W_q_out, elements_per_sample


@torch.library.custom_op("gemlite::unpack_over_rows_torch", mutates_args=())
def unpack_over_rows_torch(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_rows: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:

    num_rows, num_cols = W_q_packed.shape
    elements_per_sample = num_output_rows // num_rows

    shifts = (
        torch.arange(
            elements_per_sample, device=W_q_packed.device, dtype=W_q_packed.dtype
        )
        * W_nbits
    )
    mask = (1 << W_nbits) - 1
    W_q_unpacked = ((W_q_packed.unsqueeze(-1) >> shifts) & mask).to(dtype)
    W_q_unpacked = W_q_unpacked.permute(0, 2, 1).reshape(num_output_rows, num_cols)

    return W_q_unpacked


@torch.library.register_fake("gemlite::unpack_over_rows_torch")
def unpack_over_rows_torch_fake(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_rows: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:
    num_rows, num_cols = W_q_packed.shape
    return torch.empty(
        (num_output_rows, num_cols), dtype=dtype, device=W_q_packed.device
    )


@triton.jit
def unpack_weights_over_rows_kernel(
    W_q_packed_ptr,
    W_q_unpacked_ptr,
    num_input_rows,
    num_cols,
    num_packed_rows,
    elements_per_sample: tl.constexpr,
    W_nbits: tl.constexpr,
    unroll: tl.constexpr,
    output_dtype: tl.constexpr,
):

    pid = tl.program_id(0)
    num_blocks = tl.cdiv(num_cols, unroll)
    pid_row = pid // num_blocks
    pid_block = pid % num_blocks
    start_col = pid_block * unroll
    packed_row = pid_row // elements_per_sample

    # Load
    cols = start_col + tl.arange(0, unroll)
    offset = packed_row * num_cols + cols
    packed_values = tl.load(W_q_packed_ptr + offset)

    # Unpack
    shift = (pid_row % elements_per_sample) * W_nbits
    mask = (1 << W_nbits) - 1
    unpacked_values = ((packed_values >> shift) & mask).to(output_dtype)

    # Store
    unpacked_offsets = pid_row * num_cols + cols
    tl.store(W_q_unpacked_ptr + unpacked_offsets, unpacked_values)


@torch.library.custom_op("gemlite::unpack_over_rows_triton", mutates_args=())
def unpack_over_rows_triton(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_rows: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:

    num_packed_rows, num_cols = W_q_packed.shape
    elements_per_sample = num_output_rows // num_packed_rows

    # Allocate output
    W_q_unpacked = torch.empty(
        (num_output_rows, num_cols), dtype=dtype, device=W_q_packed.device
    )
    output_dtype = TORCH_DTYPE_TO_TRITON[dtype]

    # Define grid
    unroll = highest_divisor(num_cols, max_val=256)
    grid = (triton.cdiv(num_output_rows * num_cols, unroll),)

    # Launch kernel
    unpack_weights_over_rows_kernel[grid](
        W_q_packed.contiguous(),
        W_q_unpacked,
        num_output_rows,
        num_cols,
        num_packed_rows,
        elements_per_sample,
        W_nbits,
        unroll,
        output_dtype,
        num_stages=2,
        num_warps=1,
    )

    return W_q_unpacked


@torch.library.register_fake("gemlite::unpack_over_rows_triton")
def unpack_over_rows_triton_fake(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_rows: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:
    num_rows, num_cols = W_q_packed.shape
    return torch.empty(
        (num_output_rows, num_cols), dtype=dtype, device=W_q_packed.device
    )


############################################################################################
pack_weights_over_rows = pack_weights_over_rows_triton
pack_weights_over_cols = pack_weights_over_cols_triton
unpack_over_rows = unpack_over_rows_triton
unpack_over_cols = unpack_over_cols_triton



# Copyright 2024-2025 The Alibaba Wan Team Authors. All rights reserved.
import html
import string

import ftfy
import regex as re
from transformers import AutoTokenizer

__all__ = ["HuggingfaceTokenizer"]


def basic_clean(text):
    text = ftfy.fix_text(text)
    text = html.unescape(html.unescape(text))
    return text.strip()


def whitespace_clean(text):
    text = re.sub(r"\s+", " ", text)
    text = text.strip()
    return text


def canonicalize(text, keep_punctuation_exact_string=None):
    text = text.replace("_", " ")
    if keep_punctuation_exact_string:
        text = keep_punctuation_exact_string.join(
            part.translate(str.maketrans("", "", string.punctuation))
            for part in text.split(keep_punctuation_exact_string)
        )
    else:
        text = text.translate(str.maketrans("", "", string.punctuation))
    text = text.lower()
    text = re.sub(r"\s+", " ", text)
    return text.strip()


class HuggingfaceTokenizer:
    def __init__(self, name, seq_len=None, clean=None, **kwargs):
        assert clean in (None, "whitespace", "lower", "canonicalize")
        self.name = name
        self.seq_len = seq_len
        self.clean = clean

        # init tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(name, **kwargs)
        self.vocab_size = self.tokenizer.vocab_size

    def __call__(self, sequence, **kwargs):
        return_mask = kwargs.pop("return_mask", False)

        # arguments
        _kwargs = {"return_tensors": "pt"}
        if self.seq_len is not None:
            _kwargs.update(
                {
                    "padding": "max_length",
                    "truncation": True,
                    "max_length": self.seq_len,
                }
            )
        _kwargs.update(**kwargs)

        # tokenization
        if isinstance(sequence, str):
            sequence = [sequence]
        if self.clean:
            sequence = [self._clean(u) for u in sequence]
        ids = self.tokenizer(sequence, **_kwargs)

        # output
        if return_mask:
            return ids.input_ids, ids.attention_mask
        else:
            return ids.input_ids

    def _clean(self, text):
        if self.clean == "whitespace":
            text = whitespace_clean(basic_clean(text))
        elif self.clean == "lower":
            text = whitespace_clean(basic_clean(text)).lower()
        elif self.clean == "canonicalize":
            text = canonicalize(basic_clean(text))
        return text



from .clip import CLIPModel
from .t5 import T5EncoderModel
from .vae import WanVAE
from .wav2vec2 import Wav2Vec2Model


__all__ = [
    "CLIPModel",
    "T5EncoderModel",
    "WanVAE",
    "Wav2Vec2Model",
]
